{
********************************************************************************
*  Name    : SDFileSystem.BAS                                                  *
*  Author  : S Wright                                                          *
*  Notice  : Copyright (c) 2006 S Wright                                       *
*  Date    : 01/03/2008                                                        *
*  Version : 3.1.1                                                             *
*  Notes   : Module for reading and writing from and to MMC, SD or SDHC cards  *
*          : using the FAT16 or FAT32 file system                              *
*          :                                                                   *
*  Uses    : FSR0, FSR1                                                        *
*          :                                                                   *
*  License : Permission is hereby granted, free of charge, To any person       *
*          : obtaining a copy of this library and associated documentation     *
*          : files (the "Software"), to deal in the Software without           *
*          : restriction, including without limitation the rights to use,      *
*          : copy, modify, merge, publish, distribute, sublicense, and/or sell *
*          : copies of the Software, and to permit persons to whom the         *
*          : Software is furnished to do so, subject to the following          *
*          : conditions:                                                       *
*          :                                                                   *
*          : The above copyright notice and this permission notice shall be    *
*          : included in all copies or substantial portions of the Software.   *
*          :                                                                   *
*          : THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,   *
*          : EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES   *
*          : OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND          *
*          : NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT       *
*          : HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,      *
*          : WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING      *
*          : FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR     *
*          : OTHER DEALINGS IN THE SOFTWARE.                                   *
*          :                                                                   *
*  Revs    : 1.0.2 13/04/07 - SPI routines changed to allow for silicon        *
*          : errors for some PICS - thanks to Tom Estes                        *
*          : 1.0.3 11/09/07 - InitSequence altered to recognise inserted card  *
*          : more quickly - SeekResponse for Cmd 1 reduced To $FF              *
*          : 1.0.4 22/09/07 - SPI altered to use SSPIF rather than SSPStatus - *
*          : thanks to SteveB on forum -                                       *
*          : www.sfcompiler.co.uk/forum/viewtopic.php?t=158                    *
*          : 1.0.5 23/09/07 - SSPSTAT.7 (SMP) changed to 0 - sample at middle  *
*          : 1.0.6 21/10/07 - Calculation of last cluster corrected - effects  *
*          : writing of files when disk full, as well as DiskSizeKB and        *
*          : FreeSpaceKB - previous version overestimated size and free space  *
*          : by one cluster for some cards                                     *
*          : 1.0.7 23/10/07 - Correction to AppendFile and CloseFile to avoid  *
*          : errors when appending to files of zero length and closing them    *
*          : 2.0.0 24/10/07 - Added OpenFileRW and ability to read and write   *
*          : at file pointer, when file opened with NewFile, OpenFileRW and    *
*          : AppendFile - can use new command FSeek to move file pointer and   *
*          : FilePtr to locate position of file pointer                        *
*          : 2.0.1 27/10/07 - DiskMounted altered to return False only if disk *
*          : not present or Init not previously called - a read/write error    *
*          : no longer causes a False return                                   *
*          : 3.0.0 26/01/08 - Redone init code to include support for SDHC     *
*          : high capacity cards and added MemoryType to return type of memory *
*          : card (MMMC, SD or SDHC); included code to slow SPI bus to max     *
*          : speed of 400kHz during card initialisation                        *
*          : 3.0.1 01/02/08 - Added FATType to return card FAT type (FAT16 or  *
*          : FAT32)                                                            *
*          : 3.0.2 02/02/08 - Added SerialNumber to get the card serial number *
*          : from the CID register                                             *
*          : 3.0.3 17/02/08 - Added support for FAT32 - allows use of much     *
*          : larger card capacities                                            *
*          : 3.0.4 21/02/08 - Added ReadNBytes                                 *
*          : 3.0.5 22/02/08 - Init altered when using hardware MSSP - SPI      *
*          : speed now set using #option SD_SPI_SPEED and not passed as        *
*          : parameter to Init                                                 *
*          : 3.0.6 24/02/08 - Added Format to re-construct full FAT16/32 card  *
*          : structure and file system - FAT16 or FAT32 chosen automatically   *
*          : depending on card size                                            *
*          : 3.0.7 24/02/08 - Init return type altered to provide more info    *
*          : 3.0.8 26/02/08 - Added DestroyFormat - USE WITH CARE!             *
*          : 3.0.9 27/02/08 - Added FileCount and DirCount to return number of *
*          : files or directories in the current subdirectory                  *
*          : 3.1.0 28/02/08 - Correction to ChDir routine so that doesn't      *
*          : incorrectly handle "." and ".." entries when passed as a string   *
*          : parameter                                                         *
*          : 3.1.1 29/02/08 - Added #option SD_REINIT_SPI to allow for         *
*          : reinitialising MSSP SPI port before each use, to allow port to be *
*          : reconfigured for other devices on SPI bus between calls to SD     *
*          : commands                                                          *
*          :                                                                   *
********************************************************************************
}
Module SD

Include "System.bas"
// Include "USART.bas"
// Include "Convert.bas"

// Default module options - user options can override these values...
#option SD_CS = PORTC.2       // SPI CS To SD CS (SD pin 1)
#option SD_DI = PORTC.5       // SPI SDO to SD DI (SD Pin 2) - use 50k to 100k ohms pull-up resistor
#option SD_CLK = PORTC.3      // SPI SCK To SD CLK (SD Pin 5)
#option SD_DO = PORTC.4       // SPI SDI To SD DO (SD Pin 7) - use 50k to 100k ohms pull-up resistor
#option SD_SPI = MSSP         // Either MSSP to use hardware MSSP for SPI, or SW to use software option (free choice of pin connection)
#option SD_SPI_SPEED = spiOscDiv64          // Use when SD_SPI = MSSP - valid speeds are spiOscDiv4, spiOscDiv16 and spiOscDiv64
#option SD_SUPPORT_SUB_DIRECTORIES = TRUE
#option SD_REINIT_SPI = FALSE               // If TRUE, SPI port reinitialsed when communicating with card; allows use of SPI port for other devices also

#if IsOption(SD_SPI) And Not (SD_SPI in (SW, MSSP))
   #error SD_SPI, "Invalid option. SPI type not recognized."
#endif

#if IsOption(SD_SPI_SPEED) And Not (SD_SPI_SPEED in (spiOscDiv4, spiOscDiv16, spiOscDiv64))
   #error SD_SPI_SPEED, "Invalid option. SPI_SPEED must be spiOscDiv4, spiOscDiv16 or spiOscDiv64."
#endif

#if IsOption(SD_SPI_SPEED)
   #if SD_SPI_SPEED = spiOscDiv4
      Const SPISpeed = %00                    // SPI master mode, clock = Fosc/04
   #elseif SD_SPI_SPEED = spiOscDiv16
      Const SPISpeed = %01                    // SPI master mode, clock = Fosc/16
   #elseif SD_SPI_SPEED = spiOscDiv64
      Const SPISpeed = %10                    // SPI master mode, clock = Fosc/64
   #endif
#endif

#if IsOption(SD_REINIT_SPI) And Not (SD_REINIT_SPI in (TRUE, FALSE))
   #error SD_REINIT_SPI, "Invalid option. SD_REINIT_SPI must be TRUE or FALSE."
#endif

#if IsOption(SD_SUPPORT_SUB_DIRECTORIES) And Not (SD_SUPPORT_SUB_DIRECTORIES in (TRUE, FALSE))
   #error SD_SUPPORT_SUB_DIRECTORIES, "Invalid option. SD_SUPPORT_SUB_DIRECTORIES must be TRUE or FALSE."
#endif

#if IsOption(SD_SPI) And SD_SPI in (MSSP)
   // Map registers to SSP
   #if _mssp = 0
      #error _device + " does not support MSSP."
   // Single SSP...
   #elseif _mssp = 1
   Dim
      SSPControl1 As SSPCON1,
      SSPStatus   As SSPSTAT,
      SSPBuffer   As SSPBUF,
      SSPIF       As PIR1.3
   // Has more than one SSP module...   
   #else
   Dim                         
      SSPControl1 As SSP1CON1,   // Map to MSSP1
      SSPStatus   As SSP1STAT,
      SSPBuffer   As SSP1BUF,
      SSPIF       As PIR1.3
   #endif
#endif

// Validate CS pin...
#if IsOption(SD_CS) And Not IsValidPortPin(SD_CS) 
   #error SD_CS, "Invalid option. CS must be a valid port pin."
#endif

// Validate DI pin...
#if IsOption(SD_DI) And Not IsValidPortPin(SD_DI) 
   #error SD_DI, "Invalid option. DI must be a valid port pin."
#endif

// Validate CLK pin...
#if IsOption(SD_CLK) And Not IsValidPortPin(SD_CLK) 
   #error SD_CLK, "Invalid option. CLK must be a valid port pin."
#endif

// Validate DO pin...
#if IsOption(SD_DO) And Not IsValidPortPin(SD_DO) 
   #error SD_DO, "Invalid option. DO must be a valid port pin."
#endif

Const
   CMD0  = $40 + 0,     // GO_IDLE_STATE 
   CMD1  = $40 + 1,     // SEND_OP_COND 
   CMD8  = $40 + 8,     // SEND_IF_COND 
   CMD9  = $40 + 9,     // SEND_CSD 
   CMD10 = $40 + 10,    // SEND_CID 
   CMD12 = $40 + 12,    // STOP_TRANSMISSION 
   CMD16 = $40 + 16,    // SET_BLOCKLEN 
   CMD17 = $40 + 17,    // READ_SINGLE_BLOCK 
   CMD18 = $40 + 18,    // READ_MULTIPLE_BLOCK 
   CMD23 = $40 + 23,    // SET_BLOCK_COUNT 
   CMD24 = $40 + 24,    // WRITE_BLOCK 
   CMD25 = $40 + 25,    // WRITE_MULTIPLE_BLOCK 
   CMD41 = $40 + 41,    // SEND_OP_COND (ACMD) 
   CMD55 = $40 + 55,    // APP_CMD 
   CMD58 = $40 + 58     // READ_OCR
   
Const
   TimeOutLimit = _clock * 40             // Set timeout limit based on clock

Public Const                           
   // Dir search & Rename types...                    
   sdFile      = 0,
   sdDirectory = 1,
   // Dir search directions...                    
   dirFirst     = 0,
   dirNext      = 1,
   dirPrevious  = 2,
   // ChDir commands...                    
   cdUp         = 0,                      // Change to parent directory
   cdRoot       = 1,                      // Change to Root Directory
   // Error codes...                    
   errOK               = $00,
   errDiskFull         = $01,
   errNotFound         = $02,
   errRootDirFull      = $03,
   errDirNotEmpty      = $04,
   errExists           = $05,
   errInUse            = $06,
   errRWError          = $07,
   errFileNotOpen      = $08,
   errBeyondEOF        = $09,
   errNoResponse       = $0A,
   errInvalidFormat    = $0B,
   errInvalidPath      = $0C,
   // Card types...
   sdMMC    = 1,                          // Multi Media Card (MMC)
   sdSD     = 2,                          // Secure Digital Card (SD)
   sdSDHC   = 4,                          // High Capacity Secure Digital Card (SDHC)
   // FAT types...
   sdFAT16  = 1,
   sdFAT32  = 2   
   
Dim        
   CS As SD_CS.SD_CS@,
   DI As SD_DI.SD_DI@,
   CLK As SD_CLK.SD_CLK@,
   DO As SD_DO.SD_DO@
 
Structure TName
   Name As String(9)
   Extension As String(4)
End Structure

Public Type TSDName = TName
   
Structure TFile
   SFR As Byte
   DirectoryStartSector As LongWord
   IsRootDirectory As SFR.Booleans(0)
   FileEntryDirectorySector As LongWord
   FileEntryDirectorySectorPos As Word
   FirstCluster As LongWord
   FATSector As LongWord
   FATSectorPos As Word
   ClusterSequenceIndex As Word
   ClusterSequenceNumber As Word
   CurrentSectorInCluster As Byte
   CurrentSector As LongWord
   CurrentSectorPos As Word
   Size As LongWord
   Number As Integer
   BytesRead As LongWord
   EOF As SFR.Booleans(1)
   DirFileDir As SFR.Bits(2)
   IsOpen As SFR.Booleans(3)
   RW As SFR.Bits(4)   // 0 = Write, 1 = Read
End Structure   
     
Structure TDisk
   SectorsPerCluster As Byte
   FAT1 As LongWord
   FAT2 As LongWord
   SectorsPerFAT As LongWord
   RootDirectory As LongWord
   SectorsInRoot As Word
   Data As LongWord
   LastCluster As LongWord
   SFR As Byte
   DiskFull As SFR.Booleans(0)
   RWError As SFR.Booleans(1)
   IsDirty As SFR.Booleans(2)
   CardType As Byte
   FATType As Byte
   SlowSPI As Byte
End Structure

Structure TShared
   CurrentSectorBuffer($200) As Byte
End Structure

Dim Disk As TDisk
Dim File(2) As TFile
Dim Shared As TShared  


{
********************************************************************************
* Name    : ReceiveByte (PRIVATE)                                              *
* Purpose : Receive a byte from card using SPI interface.                      *
*         : Either software or hardware SPI can be used by selecting           *
*         : #option SD_SPI = SW Or #option SD_SPI = MSSP                       *
********************************************************************************
}   
#if SD_SPI = SW                           // Software SPI version
Function ReceiveByte() As Byte
Dim Shift As Byte 
   ReceiveByte = 0
   Shift = 8
   If Disk.SlowSPI > 0 Then 
      Repeat
         ReceiveByte = ReceiveByte << 1      // Shift in...
         ReceiveByte.0 = DO                  // Sample before clock...
         DelayUS(2)
         CLK = 1
         DelayUS(2)
         CLK = 0
         Dec(Shift)
      Until Shift = 0
   Else
      Repeat
         ReceiveByte = ReceiveByte << 1      // Shift in...
         ReceiveByte.0 = DO                  // Sample before clock...
         CLK = 1
         CLK = 0
         Dec(Shift)
      Until Shift = 0
   EndIf
   ClrWDT
End Function
#elseif SD_SPI = MSSP                     // MSSP version
Function ReceiveByte() As Byte
Dim Shift As Byte 
   If Disk.SlowSPI > 0 Then 
      ReceiveByte = 0
      Shift = 8
      Repeat
         ReceiveByte = ReceiveByte << 1   // Shift in...
         ReceiveByte.0 = DO               // Sample before clock...
         DelayUS(2)
         CLK = 1
         DelayUS(2)
         CLK = 0
         Dec(Shift)
      Until Shift = 0
      ClrWDT
   Else 
      SSPIF = 0 
      SSPBuffer = $FF 
      Repeat 
         ClrWDT 
      Until SSPIF = 1 
      ReceiveByte = SSPBuffer     
   EndIf
End Function
#endif
{
********************************************************************************
* Name    : SeekResponse (PRIVATE)                                             *
* Purpose : Wait for a specified response from card                            *
********************************************************************************
}   
Function SeekResponse(pRequired As Byte) As Byte
Dim Index As Word
   Index = $FFFF
   Repeat
      SeekResponse = ReceiveByte()
      Dec(Index)
   Until SeekResponse = pRequired Or Index = 0
End Function                                                           
{
********************************************************************************
* Name    : SendByte (PRIVATE)                                                 *
* Purpose : Send a byte to card using SPI interface.                           *
*         : Either software Or hardware SPI can be used by selecting           *
*         : #option SD_SPI = SW or #option SD_SPI = MSSP                       *
********************************************************************************
}   
#if SD_SPI = SW                  // Software SPI version
Sub SendByte(pByte As Byte)
Dim Shift As Byte 
   Shift = 8                     // Write 8 bits...
   If Disk.SlowSPI > 0 Then 
      Repeat
         CLK = 0
         DI = pByte.Bits(7)
         pByte = pByte << 1
         Dec(Shift)
         DelayUS(2)
         CLK = 1
         DelayUS(2)
      Until Shift = 0
   Else
      Repeat
         CLK = 0
         DI = pByte.Bits(7)
         pByte = pByte << 1
         Dec(Shift)
         CLK = 1
      Until Shift = 0
   EndIf
   CLK = 0                       // Clock idles low...
   ClrWDT
End Sub
#elseif SD_SPI = MSSP            // MSSP version
Sub SendByte(pByte As Byte)
Dim Shift As Byte 
   If Disk.SlowSPI > 0 Then 
      Shift = 8                  // Write 8 bits...
      Repeat
         CLK = 0
         DI = pByte.Bits(7)
         pByte = pByte << 1
         Dec(Shift)
         DelayUS(2)
         CLK = 1
         DelayUS(2)
      Until Shift = 0
      CLK = 0                    // Clock idles low...
      ClrWDT
   Else
      SSPIF = 0 
      SSPBuffer = pByte 
      Repeat 
         ClrWDT 
      Until SSPIF = 1 
   EndIf
End Sub
#endif
{
********************************************************************************
* Name    : SendCmd (PRIVATE)                                                  *
* Purpose : Send a command to card using SPI interface                         *
********************************************************************************
}   
Function SendCmd(pCmd As Byte, pArg As LongWord) As Byte
Dim CRC, Index As Byte	
   CS = 1                     // Select card and wait for ready
   CS = 0
	If SeekResponse($FF) <> $FF Then   // Wait for ready
      SendCmd = $FF
   Else 
   	SendByte(pCmd)          // Command
   	SendByte(pArg.Byte3)    // Argument[31..24]
   	SendByte(pArg.Byte2)    // Argument[23..16]
   	SendByte(pArg.Byte1)    // Argument[15..08]
   	SendByte(pArg.Byte0)    // Argument[07..00]
   	CRC = $FF				   // Default CRC byte
   	If pCmd = CMD0 Then
         CRC = $95			   // CRC for CMD0($0)
      ElseIf pCmd = CMD8 Then
         CRC = $87			   // CRC for CMD8($1AA)
      EndIf
   	SendByte(CRC)
   	Index = 100
   	Repeat                                     // Receive command response
   		Dec(Index)
         SendCmd = ReceiveByte()
   	Until (SendCmd And $80) = 0 Or Index = 0   // Wait for a valid response in timeout of 100 attempts
  	EndIf
End Function
{
********************************************************************************
* Name    : ReInitSPIPort (PRIVATE)                                            *
* Purpose : Reinitialises SPI for communication with card; allows other        *
*         : devices to use SPI port also                                       *
********************************************************************************
}   
Sub ReInitSPIPort()
   SSPControl1 = SPISpeed Or %00100000   // Setup MSSP module & turn on
   SSPStatus = %01000000                         
End Sub
{
********************************************************************************
* Name    : ReadSector (PRIVATE)                                               *
* Purpose : Read a full sector from card to buffer.                            *
********************************************************************************
}   
Sub ReadSector(pSector As LongWord, pBuffer As Boolean = True)
Dim TimeOut, Error As Byte
Dim Index As Word
#if SD_REINIT_SPI = TRUE
   ReInitSPIPort()
#endif
   If (Disk.RWError = True And pBuffer) Then     // If RW error has occurred, prevent further reading from 
                                                 // disk until reinitialised, unless not buffering
                                                 // (pBuffer = False, used for DiskMounted)
      Exit
   EndIf
   TimeOut = $00
   Error = 1
   If (Disk.CardType And $04) = 0 Then           // MMC or SD - byte addressing (SDHC use block addressing)
      pSector = pSector << 9
   EndIf 
   Repeat
      CS = 0                                     
      If SendCmd(CMD17, pSector) = $00 Then      // Send Cmd 17
         If SeekResponse($FE) = $FE Then         // Read start token
            Index = 0
            If pBuffer Then
               FSR0 = @Shared.CurrentSectorBuffer
               Repeat                            // Read data block
#if SD_SPI = SW                                  // Software SPI version - use SendByte
                  POSTINC0 = ReceiveByte()
#elseif SD_SPI = MSSP                            // MSSP version - use inline routine to optimise speed
                  SSPIF = 0 
                  SSPBuffer = $FF 
                  Repeat 
                     ClrWDT 
                  Until SSPIF = 1 
                  POSTINC0 = SSPBuffer
#endif
                  Inc(Index)
               Until Index = $200
            Else
               Repeat                            // Don't read data block
                  ReceiveByte()
                  Inc(Index)
               Until Index = $200
            EndIf
            ReceiveByte()                        // Read dummy CRC to conclude data block
            ReceiveByte()
            Error = 0                            // Read completed successfully
         EndIf   
      EndIf      
      CS = 1
      SendByte($FF)                              // Clock SD/MMC to complete read
      Inc(TimeOut)
   Until Error = 0 Or TimeOut > $01
   If Error = 1 Then
      Disk.RWError = True                        // Set RWError flag - can only be cleared using Init()
   EndIf
End Sub                                                                
{
********************************************************************************
* Name    : SectorInitRead (PRIVATE)                                           *
* Purpose : Initialise sector for reading.                                     *
********************************************************************************
}   
Sub SectorInitRead(pSector As LongWord)
   File(1).CurrentSectorPos = 0               
   ReadSector(pSector)
End Sub         
{
********************************************************************************
* Name    : SectorInitWrite (PRIVATE)                                          *
* Purpose : Initialise sector for writing.                                     *
********************************************************************************
}   
Sub SectorInitWrite()
   File(1).CurrentSectorPos = 0               
   Clear(Shared.CurrentSectorBuffer)
End Sub         
{
********************************************************************************
* Name    : WriteSector (PRIVATE)                                              *
* Purpose : Write a full sector from buffer to card                            *
********************************************************************************
}   
Sub WriteSector(pSector As LongWord)
Dim Response As Byte
Dim Index As Word 
Dim TimeOut As Byte
Dim Error As Byte
#if SD_REINIT_SPI = TRUE
   ReInitSPIPort()
#endif
   If Disk.RWError = True Then           // If RW error has occurred, prevent further writing to disk until
                                         // reinitialised
      Exit
   EndIf
   TimeOut = $00
   Error = 1
   If (Disk.CardType And $04) = 0 Then   // MMC or SD - byte addressing (SDHC use block addressing)
      pSector = pSector << 9
   EndIf 
   Repeat
      CS = 0                             // Send Cmd 24
      If SeekResponse($FF) = $FF Then    // Wait for ready
         If SendCmd(CMD24, pSector) = $00 Then
            SendByte($FE)                // Write start token 
            Index = 0
            FSR0 = @Shared.CurrentSectorBuffer
            Repeat                       // Write data block
#if SD_SPI = SW                          // Software SPI version - use SendByte
               SendByte(POSTINC0)
#elseif SD_SPI = MSSP                    // MSSP version - use inline routine to optimise speed
               SSPIF = 0 
               SSPBuffer = POSTINC0 
               Repeat 
                  ClrWDT 
               Until SSPIF = 1 
#endif                        
               Inc(Index)
            Until Index = $200           
            SendByte($FF)                // Send dummy CRC to conclude data block
            SendByte($FF)                                                           
            Response = ReceiveByte()     // Data response token
            Response = Response And $1F
            If Response = $05 Then
               Error = 0
            EndIf
         EndIf
      EndIf
      CS = 1
      Inc(TimeOut)
   Until Error = 0 Or TimeOut > 10
   If Error = 1 Then
      Disk.RWError = True                // Set RWError flag - can only be cleared using Init()
   EndIf
End Sub
{
********************************************************************************
* Name    : WriteSectorByte (PRIVATE)                                          *
* Purpose : Write a byte to current sector.                                    *
*         : Can be used with either Sector or FA16 subroutines                 *
********************************************************************************
}   
Sub WriteSectorByte(pByte As Byte)
   If File(1).CurrentSectorPos = $200 Then
      File(1).CurrentSectorPos = $000
   EndIf
   Shared.CurrentSectorBuffer(File(1).CurrentSectorPos) = pByte
   Disk.IsDirty = True
   If File(1).CurrentSectorPos = $1FF Then
      WriteSector(File(1).CurrentSector)
      Disk.IsDirty = False
      Inc(File(1).CurrentSector)
   EndIf
   Inc(File(1).CurrentSectorPos)
End Sub
{
********************************************************************************
* Name    : ReadFATEntry (PRIVATE)                                             *
* Purpose : Read an entry from the FAT                                         *
********************************************************************************
}   
Function ReadFATEntry(pSectorPos As Word) As LongWord                           
   ReadFATEntry.Byte0 = Shared.CurrentSectorBuffer(pSectorPos)                  // Read FAT entry as longword
   ReadFATEntry.Byte1 = Shared.CurrentSectorBuffer(pSectorPos + 1)
   If Disk.FATType = 1 Then   // FAT16
      ReadFATEntry.Byte2 = 0
      ReadFATEntry.Byte3 = 0      
   Else                       // FAT32
      ReadFATEntry.Byte2 = Shared.CurrentSectorBuffer(pSectorPos + 2)
      ReadFATEntry.Byte3 = Shared.CurrentSectorBuffer(pSectorPos + 3) And $0F   // Clear high 4 bits for 28 bit FAT entry  
   EndIf
End Function
{
********************************************************************************
* Name    : FATSectorToCluster (PRIVATE)                                       *
* Purpose : Calculate cluster number based on FAT sector/position              *
********************************************************************************
}   
Function FATSectorToCluster() As LongWord                    // FAT16 & FAT32   
   FATSectorToCluster = File(1).FATSectorPos / (2 * Disk.FATType)
   FATSectorToCluster = FATSectorToCluster + (File(1).FATSector * $100 / Disk.FATType)
End Function
{
********************************************************************************
* Name    : ClusterToFATSector (PRIVATE)                                       *
* Purpose : Calculate FAT Sector/position based on cluster Number              *
********************************************************************************
}   
Sub ClusterToFATSector(pCluster As LongWord)                                    
   If Disk.FATType = 1 Then   // FAT16
      File(1).FATSector = pCluster / $100   // Find FAT position for existing cluster
      File(1).FATSectorPos = pCluster Mod $100
      File(1).FATSectorPos = File(1).FATSectorPos * 2
   Else                       // FAT32
      File(1).FATSector = pCluster / $80   // Find FAT position for existing cluster
      File(1).FATSectorPos = pCluster Mod $80
      File(1).FATSectorPos = File(1).FATSectorPos * 4
   EndIf 
End Sub
{
********************************************************************************
* Name    : DataToCluster (PRIVATE)                                            *
* Purpose : Calculate cluster number based on data sector                      *
********************************************************************************
}   
Function DataToCluster() As LongWord                         // FAT16 & FAT32   
Dim TempCluster As LongWord
   TempCluster = File(1).CurrentSector - Disk.Data       
   TempCluster = TempCluster / Disk.SectorsPerCluster
   DataToCluster = TempCluster + 2
End Function
{
********************************************************************************
* Name    : ClusterToData (PRIVATE)                                            *
* Purpose : Calculate data sector based on cluster number                      *
********************************************************************************
}   
Function ClusterToData(pCluster As LongWord) As LongWord     // FAT16 & FAT32   
   ClusterToData = (pCluster - 2) * Disk.SectorsPerCluster
   ClusterToData = ClusterToData + Disk.Data
End Function
{
********************************************************************************
* Name    : FindFreeCluster (PRIVATE)                                          *
* Purpose : Find a free cluster in the FAT                                     *
********************************************************************************
}   
Function FindFreeCluster() As Byte                                              
Dim ScanSector As LongWord
Dim ScanSectorPos As Word
Dim FATEntry As LongWord
Dim Cluster As LongWord
   FindFreeCluster = 1                          // Scan FAT1 for free cluster address
   ScanSector = (Disk.FAT1 + File(1).FATSector)
   Repeat
      SectorInitRead(ScanSector)                // Set sector & read sector to buffer
      If ScanSector = (Disk.FAT1 + File(1).FATSector) Then
         ScanSectorPos = File(1).FATSectorPos + (2 * Disk.FATType * File(1).ClusterSequenceNumber)
      Else
         ScanSectorPos = $000
      EndIf
      Repeat
         FATEntry = ReadFATEntry(ScanSectorPos)
         If FATEntry = $00000000 Then
            FindFreeCluster = 0
            Break
         EndIf
         Inc(ScanSectorPos, 2 * Disk.FATType)
      Until ScanSectorPos > $1FF
      Inc(ScanSector)
   Until ScanSector = (Disk.FAT1 + Disk.SectorsPerFAT) Or FindFreeCluster = 0
   Dec(ScanSector)
   If FindFreeCluster = 1 Then
      Exit
   EndIf
   Cluster = ScanSectorPos / (2 * Disk.FATType)
   Cluster = Cluster + ((ScanSector - Disk.FAT1) * ($100 / Disk.FATType)) 
   If Cluster > Disk.LastCluster Then           // Check resulting cluster (hence sector)
                                                // not greater than number on card
      FindFreeCluster = 1
   Else
      File(1).FATSector = ScanSector - Disk.FAT1   // Store FAT sector
      File(1).FATSectorPos = ScanSectorPos         // Store buffer position
      File(1).ClusterSequenceNumber = 0
      ScanSectorPos = File(1).FATSectorPos
      Repeat
         FATEntry = ReadFATEntry(ScanSectorPos)
         If FATEntry = $00000000 Then
            Inc(File(1).ClusterSequenceNumber)     // Number of empty clusters in sector
         Else
            Break
         EndIf
         Inc(ScanSectorPos, 2 * Disk.FATType)
      Until ScanSectorPos > $1FF
      File(1).ClusterSequenceIndex = 1             // Initilise cluster sequence
      If (Cluster + File(1).ClusterSequenceNumber - 1) > Disk.LastCluster Then
         File(1).ClusterSequenceNumber = Disk.LastCluster - Cluster + 1
      EndIf
   EndIf
End Function
{
********************************************************************************
* Name    : MakeUpper (PRIVATE)                                                *
* Purpose : Converts single char to uppercase                                  *
********************************************************************************
}  
Function MakeUpper(pValue As Byte) As Byte
   If Char(pValue) >= "a" And Char(pValue) <= "z" Then
      Dec(pValue, 32)
   EndIf 
   Result = pValue
End Function
{
********************************************************************************
* Name    : ModifyFileName (PRIVATE)                                           *
* Purpose : Replace blank file name characters with $20                        *
********************************************************************************
}     
Sub ModifyFileName(ByRef pFileName As TSDName)
Dim RootDirFilePos As Byte
Dim NullFl As Byte
   NullFl = 0
   RootDirFilePos = $00
   Repeat
      If (NullFl = 1) Or (Byte(pFileName.Name(RootDirFilePos)) = 0) Then
         pFileName.Name(RootDirFilePos) = $20
         NullFl = 1
      Else
         pFileName.Name(RootDirFilePos) = MakeUpper(pFileName.Name(RootDirFilePos))
      EndIf
      Inc(RootDirFilePos)                            
   Until RootDirFilePos > $07
   pFileName.Name(8) = 0
   NullFl = 0
   RootDirFilePos = $00
   Repeat
      If (NullFl = 1) Or (Byte(pFileName.Extension(RootDirFilePos)) = 0) Then
         pFileName.Extension(RootDirFilePos) = $20
         NullFl = 1
      Else
         pFileName.Extension(RootDirFilePos) = MakeUpper(pFileName.Extension(RootDirFilePos))
      EndIf
      Inc(RootDirFilePos)                                                               
   Until RootDirFilePos > $02 
   pFileName.Extension(3) = 0
End Sub
{
********************************************************************************
* Name    : IncrementSector (PRIVATE)                                          *
* Purpose : Finds next sector in FAT chain.                                    *
*         : Return: 0 = Success, 1 = End of cluster chain reached              *
********************************************************************************
}   
Function IncrementSector() As Byte                                              
Dim Cluster As LongWord
   Inc(File(1).CurrentSectorInCluster)
   If File(1).CurrentSectorInCluster > Disk.SectorsPerCluster Then
      File(1).CurrentSector = File(1).CurrentSector - (Disk.SectorsPerCluster - 1)
      Cluster = DataToCluster()
      ClusterToFATSector(Cluster)     // Find FAT position for existing cluster
      File(1).CurrentSector = Disk.FAT1 + File(1).FATSector
      ReadSector(File(1).CurrentSector)
      Cluster = ReadFATEntry(File(1).FATSectorPos)
      If (Disk.FATType = 1 And Cluster = $FFFF) Or (Disk.FATType = 2 And Cluster = $0FFFFFFF) Then 
         IncrementSector = 1
      Else
         File(1).CurrentSector = ClusterToData(Cluster)
         File(1).CurrentSectorInCluster = 1
         IncrementSector = 0
         ClusterToFATSector(Cluster)  // Update FAT positions   
      EndIf
   Else
      Inc(File(1).CurrentSector)
      IncrementSector = 0
   EndIf
End Function
{
********************************************************************************
* Name    : RootDirPosition (PRIVATE)                                          *
* Purpose : Returns a byte from position pPos in Root Directory entry          *
********************************************************************************
}   
Function RootDirPosition(pPos As Byte) As Byte
   RootDirPosition = Shared.CurrentSectorBuffer(File(1).CurrentSectorPos + pPos)
End Function
{
********************************************************************************
* Name    : CheckFileEntryStatus (PRIVATE)                                     *
* Purpose : Check type & status of entry in directory.                         *
*         : Returns true if entry is a file or directory, set by pFileDir, and *
*         : not a deleted file, long filename, system, hidden or volume entry  *
********************************************************************************
}   
Function CheckFileEntryStatus(pFileDir As Bit) As Boolean
Dim Byte00 As Byte
Dim Byte0B As Byte
   Byte00 = RootDirPosition($00)
   Byte0B = RootDirPosition($0B)
   CheckFileEntryStatus = False
   If (Byte0B Or %11110000) <> $FF Then                 // Not long filename entry
      If Byte00 <> $00 Then                             // Not empty entry
         If Byte00 <> $E5 Then                          // Not deleted entry
            If (Byte0B And %00001110) = 0 Then          // Not system, hidden or volume
                                                        // (File attributes (00ADVSHR)
                                                        // 0: unused bit, A: archive bit, D: Dir bit, 
                                                        // V: volume bit, S: system bit, R: read-only bit)
               If (Byte0B.Bit4 Xor pFileDir) = 0 Then   // Check if file or directory
                  CheckFileEntryStatus = True
               EndIf
            EndIf
         EndIf
      EndIf
   EndIf
End Function
{
********************************************************************************
* Name    : FindRootDirEntry (PRIVATE)                                         *
* Purpose : Search for free root directory entry, search for file,             *
*         : or search for files using Dir.                                     *
*         : pType: 0 = Free position, 1 = Existing File/Dir, 2 = Dir command   *
*         : pFileDir: 0 = Find file, 1 = Find directory                        *
*         : Return: True = Success, False = Fail                               *
********************************************************************************
}   
Function FindRootDirEntry(ByRef pFileName As TSDName, pType As Byte, pFileDir As Bit = 0) As Boolean   
Dim RootDirFilePos As Byte
Dim RootDirFileNum As Word
Dim NullFl As Byte
Dim Cluster As LongWord
Dim Response As Byte
   ModifyFileName(pFileName)
   RootDirFileNum = 0
   FindRootDirEntry = False
   File(1).CurrentSector = File(1).DirectoryStartSector
   File(1).CurrentSectorInCluster = 1
   Repeat
      If Disk.RWError Then
         FindRootDirEntry = False
         Exit
      EndIf
      SectorInitRead(File(1).CurrentSector)                       // Read sector to buffer
      File(1).CurrentSectorPos = $000
      Repeat
         Select pType
         Case 0                                                   // Search for free position
            If RootDirPosition(0) = $00 Or RootDirPosition(0) = $E5 Then
               FindRootDirEntry = True
            EndIf
         Case 1                                                   // Search for file or directory by name
            If RootDirPosition(0) = $00 Then                      // Once found first $00 entry, must have reached end of valid files in directory
               Exit
            EndIf
            NullFl = 0
            RootDirFilePos = $00
            Repeat                                                // Check file/dir name
               If RootDirPosition(RootDirFilePos) <> Byte(pFileName.Name(RootDirFilePos)) Then 
                  NullFl = 1
                  Break
               EndIf
               Inc(RootDirFilePos)
            Until RootDirFilePos > $07
            RootDirFilePos = $08
            Repeat                                                // Check file/dir extension
               If RootDirPosition(RootDirFilePos) <> Byte(pFileName.Extension(RootDirFilePos - $08)) Then 
                  NullFl = 1
                  Break
               EndIf
               Inc(RootDirFilePos)
            Until RootDirFilePos > $0A
            If NullFl = 0 Then
               If CheckFileEntryStatus(pFileDir) Then
                  Cluster.Byte0 = RootDirPosition($1A)            // Locate cluster number
                  Cluster.Byte1 = RootDirPosition($1B)
                  Cluster.Byte2 = RootDirPosition($14)            // High byte for FAT32
                  Cluster.Byte3 = RootDirPosition($15)
                  File(1).Size.Byte0 = RootDirPosition($1C)       // Obtain file size (bytes)
                  File(1).Size.Byte1 = RootDirPosition($1D)
                  File(1).Size.Byte2 = RootDirPosition($1E)
                  File(1).Size.Byte3 = RootDirPosition($1F)
                  FindRootDirEntry = True
               EndIf
            EndIf
         Case 2                                                   // Dir command
            If RootDirPosition(0) = $00 Then                      // Once found first $00 entry, must have reached end of valid files in directory
               Exit
            EndIf
            If CheckFileEntryStatus(pFileDir) Then
               Inc(RootDirFileNum)
               If RootDirFileNum = File(1).Number Then            // Found required entry in directory
                  RootDirFilePos = $00
                  Repeat                                          // Transfer file name
                     pFileName.Name(RootDirFilePos) = RootDirPosition(RootDirFilePos)
                     Inc(RootDirFilePos)
                  Until RootDirFilePos > $07
                  pFileName.Name($08) = 0                         // String terminator
                  Repeat                                          // Transfer file extension
                     pFileName.Extension(RootDirFilePos - $08) = RootDirPosition(RootDirFilePos) 
                     Inc(RootDirFilePos)
                  Until RootDirFilePos > $0A
                  pFileName.Extension($03) = 0                    // String terminator
                  FindRootDirEntry = True
               EndIf
            EndIf
         EndSelect
         Inc(File(1).CurrentSectorPos, 32)
      Until (File(1).CurrentSectorPos > $1FF) Or FindRootDirEntry
      Dec(File(1).CurrentSectorPos, 32)
      If Not(FindRootDirEntry) Then
         If File(1).IsRootDirectory Then
            Inc(File(1).CurrentSector)
            If File(1).CurrentSector < (Disk.RootDirectory + Disk.SectorsInRoot) Then
               Response = 0
            Else
               Response = 1
            EndIf
         Else
            Response = IncrementSector()
         EndIf
      EndIf
   Until FindRootDirEntry Or (Response = 1)
   If Not(FindRootDirEntry) Then
      Exit
   EndIf
   File(1).FileEntryDirectorySector = File(1).CurrentSector       // Store root directory position
   File(1).FileEntryDirectorySectorPos = File(1).CurrentSectorPos // Store buffer position
   If pType = 1 Then                                              // Only perform if searching for specific entry
      If Cluster = 0 Then                                         // Only true when using ChDir to return to Root Directory using cdUp
         File(1).CurrentSector = Disk.RootDirectory
      Else
         File(1).CurrentSector = ClusterToData(Cluster)
      EndIf
   EndIf
End Function
{
********************************************************************************
* Name    : InsertByteIntoBuffer (PRIVATE)                                     *
* Purpose : Insert a byte into buffer                                          *
********************************************************************************
}   
Sub InsertByteIntoBuffer(pSectorPos As Word, pByte As Byte)
   Shared.CurrentSectorBuffer(pSectorPos) = pByte
End Sub
{
********************************************************************************
* Name    : WriteFATEntry (PRIVATE)                                            *
* Purpose : Write an entry into the FAT                                        *
********************************************************************************
}   
Sub WriteFATEntry(pCluster As LongWord, pFATSectorPos As Word) 
   InsertByteIntoBuffer(pFATSectorPos, pCluster.Byte0)
   Inc(pFATSectorPos)
   InsertByteIntoBuffer(pFATSectorPos, pCluster.Byte1)
   Inc(pFATSectorPos)
   If Disk.FATType = 2 Then                          // FAT32
      InsertByteIntoBuffer(pFATSectorPos, pCluster.Byte2)
      Inc(pFATSectorPos)
      InsertByteIntoBuffer(pFATSectorPos, (Shared.CurrentSectorBuffer(pFATSectorPos) And $F0) Or pCluster.Byte3)   // Keep high 4 bits of FAT cluster and Or with high byte of new cluster (28 bit cluster address)
      Inc(pFATSectorPos)
   EndIf
End Sub
{
********************************************************************************
* Name    : AssembleFATSector (PRIVATE)                                        *
* Purpose : Update FAT sector after cluster sequence written.                  *
*         : pType: 0 = End continue, 1 = End close                             *
********************************************************************************
}   
Sub AssembleFATSector(pFATSectorPrevious As LongWord, pFATSectorPosPrevious As Word,   
                      pClusterSequenceNumberPrevious As Word, pType As Byte)
Dim Cluster As LongWord
   Cluster = pFATSectorPrevious * $100 / Disk.FATType   // Derive first cluster number
   Cluster = Cluster + (pFATSectorPosPrevious / (2 * Disk.FATType))
   Inc(Cluster)
   While pClusterSequenceNumberPrevious > 1
      WriteFATEntry(Cluster, pFATSectorPosPrevious)
      Inc(pFATSectorPosPrevious, 2)
      If Disk.FATType = 2 Then                          // FAT32
         Inc(pFATSectorPosPrevious, 2)
      EndIf
      Inc(Cluster)
      Dec(pClusterSequenceNumberPrevious)
   Wend
   If pType = 0 Then
      WriteFATEntry(FATSectorToCluster(), pFATSectorPosPrevious)
   Else
      If Disk.FATType = 1 Then
         WriteFATEntry($FFFF, pFATSectorPosPrevious)        // FAT16
      Else
         WriteFATEntry($0FFFFFFF, pFATSectorPosPrevious)    // FAT32 
      EndIf
   EndIf
End Sub
{
********************************************************************************
* Name    : WriteByte (PRIVATE)                                                *
* Purpose : Write a single byte to file.                                       *
*         : Disk.DiskFull = True when disk full                                *
********************************************************************************
}   
Sub WriteByte(pByte As Byte)                                                    
Dim FATSectorPrevious As LongWord
Dim FATSectorPosPrevious As Word
Dim ClusterSequenceNumberPrevious As Word
Dim Cluster As LongWord
Dim Response As Byte
   If Disk.DiskFull Then                                       // Disk full
      Exit
   EndIf
   Inc(File(1).BytesRead)
   If File(1).EOF Or (File(1).BytesRead > File(1).Size) Then   // Extending file size
      Inc(File(1).Size)
      File(1).EOF = True
   EndIf
   If File(1).CurrentSectorPos > $1FF Then
      If File(1).EOF = False Then                              // Not extending file
         Dec(File(1).CurrentSector)                            // Put current sector back by one to undo
                                                               // inc on last WriteSectorByte
         IncrementSector()
      Else                                                     // Extending file
         Inc(File(1).CurrentSectorInCluster)
         If File(1).CurrentSectorInCluster > Disk.SectorsPerCluster Then
            Inc(File(1).ClusterSequenceIndex)
            If File(1).ClusterSequenceIndex > File(1).ClusterSequenceNumber Then
               FATSectorPrevious = File(1).FATSector
               FATSectorPosPrevious = File(1).FATSectorPos                      
               ClusterSequenceNumberPrevious = File(1).ClusterSequenceNumber
               Response = FindFreeCluster()
               If Response = 1 Then
                  Dec(File(1).ClusterSequenceIndex)
                  File(1).FATSector = FATSectorPrevious
                  File(1).FATSectorPos = FATSectorPosPrevious                     
                  File(1).ClusterSequenceNumber = ClusterSequenceNumberPrevious
                  Disk.DiskFull = True
                  Dec(File(1).Size)
                  Dec(File(1).BytesRead)
                  Exit
               EndIf
               ReadSector(Disk.FAT1 + FATSectorPrevious)
               AssembleFATSector(FATSectorPrevious, FATSectorPosPrevious, ClusterSequenceNumberPrevious, 0)
               WriteSector(Disk.FAT1 + FATSectorPrevious)      // Write Disk.FAT1
               WriteSector(Disk.FAT2 + FATSectorPrevious)      // Write Disk.FAT1
               Cluster = FATSectorToCluster()
               File(1).CurrentSector = ClusterToData(Cluster)
               SectorInitWrite()
            EndIf
            File(1).CurrentSectorInCluster = 1
         EndIf
      EndIf
      ReadSector(File(1).CurrentSector)
   EndIf
   WriteSectorByte(pByte)
End Sub        
{
********************************************************************************
* Name    : DiskFull (PUBLIC)                                                  *
* Purpose : Determines if a disk is full when writing to file.                 *
*         : Return: False = Disk not yet full, True = Disk is full             *
********************************************************************************
}   
Public Function DiskFull() As Boolean
   Result = Disk.DiskFull
End Function
{
********************************************************************************
* Name    : WriteItem (OVERLOAD)                                               *
* Purpose : Write a Boolean to file.                                           *
*         : Check DiskFull for disk full                                       *
********************************************************************************
}  
Sub WriteItem(pValue As Boolean)
   WriteByte(Byte(pValue))
End Sub
{
********************************************************************************
* Name    : WriteItem (OVERLOAD)                                               *
* Purpose : Write a Bit to file.                                               *
*         : Check DiskFull for disk full                                       *
********************************************************************************
}  
Sub WriteItem(pValue As Bit)
   WriteByte(Byte(pValue))
End Sub
{
********************************************************************************
* Name    : WriteItem (OVERLOAD)                                               *
* Purpose : Write a Byte to file.                                              *
*         : Check DiskFull for disk full                                       *
********************************************************************************
}  
Sub WriteItem(pValue As Byte)
   WriteByte(pValue)
End Sub
{
********************************************************************************
* Name    : WriteItem (OVERLOAD)                                               *
* Purpose : Write a Char to file.                                              *
*         : Check DiskFull for disk full                                       *
********************************************************************************
}  
Sub WriteItem(pValue As Char)
   WriteByte(pValue)
End Sub
{
********************************************************************************
* Name    : WriteItem (OVERLOAD)                                               *
* Purpose : Write a ShortInt to file.                                          *
*         : Check DiskFull for disk full                                       *
********************************************************************************
}  
Sub WriteItem(pValue As ShortInt)
   WriteByte(pValue)
End Sub
{
********************************************************************************
* Name    : WriteItem (OVERLOAD)                                               *
* Purpose : Write a Word to file.                                              *
*         : Check DiskFull For Disk full                                       *
********************************************************************************
}  
Sub WriteItem(pValue As Word)
   WriteByte(pValue.Byte0)
   WriteByte(pValue.Byte1)
End Sub
{
********************************************************************************
* Name    : WriteItem (OVERLOAD)                                               *
* Purpose : Write an Integer to file.                                          *
*         : Check DiskFull For Disk full                                       *
********************************************************************************
}  
Sub WriteItem(pValue As Integer)
   WriteByte(pValue.Byte0)
   WriteByte(pValue.Byte1)
End Sub
{
********************************************************************************
* Name    : WriteItem (OVERLOAD)                                               *
* Purpose : Write a LongWord to file.                                          *
*         : Check DiskFull For Disk full                                       *
********************************************************************************
}  
Sub WriteItem(pValue As LongWord)
   WriteByte(pValue.Byte0)
   WriteByte(pValue.Byte1)
   WriteByte(pValue.Byte2)
   WriteByte(pValue.Byte3)
End Sub
{
********************************************************************************
* Name    : WriteItem (OVERLOAD)                                               *
* Purpose : Write a LongInt to file.                                           *
*         : Check DiskFull For Disk full                                       *
********************************************************************************
}  
Sub WriteItem(pValue As LongInt)
   WriteByte(pValue.Byte0)
   WriteByte(pValue.Byte1)
   WriteByte(pValue.Byte2)
   WriteByte(pValue.Byte3)
End Sub
{
********************************************************************************
* Name    : WriteItem (OVERLOAD)                                               *
* Purpose : Write a Float to file.                                             *
*         : Check DiskFull For Disk full                                       *
********************************************************************************
}  
Sub WriteItem(pValue As Float)
   WriteByte(pValue.Byte0)
   WriteByte(pValue.Byte1)
   WriteByte(pValue.Byte2)
   WriteByte(pValue.Byte3)
End Sub
{
********************************************************************************
* Name    : WriteItem (OVERLOAD)                                               *
* Purpose : Write a String to file.                                            *
*         : Check DiskFull For Disk full                                       *
********************************************************************************
}  
Sub WriteItem(pText As String)
Dim StringPos As Byte
   StringPos = 0
   While pText(StringPos) <> null
      WriteByte(pText(StringPos))
      Inc(StringPos)
   Wend
End Sub
{
********************************************************************************
* Name    : Write (COMPOUND)                                                   *
* Purpose : Write one or more items to file.                                   *
*         : Check DiskFull For Disk full                                       *
********************************************************************************
} 
Public Compound Sub Write(WriteItem)
{
********************************************************************************
* Name    : WriteString (PUBLIC)                                               *
* Purpose : Write a String to file, passing string ByRef.                      *
*         : Check DiskFull For Disk full                                       *
********************************************************************************
}  
Public Sub WriteString(ByRef pText As String)
Dim StringPos As Byte
   StringPos = 0
   While pText(StringPos) <> null
      WriteByte(pText(StringPos))
      Inc(StringPos)
   Wend
End Sub
{
********************************************************************************
* Name    : StrToFileDirName (PRIVATE)                                         *
* Purpose : Converts a file or directory string (e.g. FILE0001.TXT) to a       *
*         : TSDName type.                                                      *
*         : String may be upto 8 characters, a '.' and 3 characters, for a     *
*         : filename, or 8 characters for a directory                          *
********************************************************************************
}   
Function StrToFileDirName(pName As String) As TSDName
Const Period = $2E
Dim Index As Byte
   StrToFileDirName.Extension = Null
   FSR1 = AddressOf(pName)  
   If INDF1 = Period Then   // Handle special cases of "." or ".." first so that not incorrectly dealt with by following routine looking for "." in filename
      StrToFileDirName.Name = pName
   Else
      StrToFileDirName.Name = Null
      Index = 0
      FSR0 = AddressOf(StrToFileDirName.Name)
      FSR1 = AddressOf(pName)  
      Index = 0                // Copy name...
      While (Index < 8) And (INDF1 <> 0) And (INDF1 <> Period)
         POSTINC0 = POSTINC1     
   	   Inc(Index)
      Wend 
      INDF0 = 0       
      If INDF1 = Period Then   // Copy extension...
         Inc(FSR1)
         Index = 0
         FSR0 = AddressOf(StrToFileDirName.Extension)
         While (Index < 3) And (INDF1 <> 0)
            POSTINC0 = POSTINC1
            Inc(Index)
         Wend   
         INDF0 = 0
      EndIf
   EndIf
End Function
{
********************************************************************************
* Name    : FileExists (OVERLOAD)                                              *
* Purpose : Set pFileName.Name & pFileName.Extension, then check for file.     *
*         : Return: True = File found, False = File not found                  *
********************************************************************************
}   
Public Function FileExists(pFileName As TSDName) As Boolean
   FileExists = FindRootDirEntry(pFileName, 1, 0)
End Function
{
********************************************************************************
* Name    : FileExists (OVERLOAD)                                              *
* Purpose : Pass pFileName As a String of upto 8 characters, a '.' and 3       *
*         : characters.                                                        *
*         : Return: True = File found, False = File not found                  *
********************************************************************************
}   
Public Function FileExists(pFileName As String) As Boolean
   Result = FileExists(StrToFileDirName(pFileName))
End Function
{
********************************************************************************
* Name    : DirExists (OVERLOAD)                                               *
* Purpose : Set pDirName.Name, then check for directory.                       *
*         : Return: True = Dir found, False = Dir not found                    *
********************************************************************************
}   
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then
Public Function DirExists(pDirName As TSDName) As Boolean
   pDirName.Extension = ""
   DirExists = FindRootDirEntry(pDirName, 1, 1)
End Function
#endif
{
********************************************************************************
* Name    : DirExists (OVERLOAD)                                               *
* Purpose : Pass pDirName as a string of 8 uppercase characters.               *
*         : Return: True = Dir found, False = Dir not found                    *
********************************************************************************
}   
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then
Public Function DirExists(pDirName As String) As Boolean
   Result = DirExists(StrToFileDirName(pDirName))
End Function
#endif
{
********************************************************************************
* Name    : ReadSectorByte (PRIVATE)                                           *
* Purpose : Read a byte from current sector.                                   *
*         : Can be used with either Sector Or FA16 subroutines                 *
********************************************************************************
}   
Function ReadSectorByte() As Byte
   If File(1).CurrentSectorPos > $1FF Then   // Inc sector if gone past end of previous sector
      Inc(File(1).CurrentSector)
      SectorInitRead(File(1).CurrentSector)                                                     
   EndIf
   ReadSectorByte = Shared.CurrentSectorBuffer(File(1).CurrentSectorPos)
   Inc(File(1).CurrentSectorPos)
End Function
{
********************************************************************************
* Name    : SetFileTimeModified (PRIVATE)                                      *
* Purpose : Set modified time & date for file                                  *
********************************************************************************
}   
Sub SetFileTimeModified(pSectorPos As Word, pDay, pMonth, pYear, pHours, pMinutes, pSeconds As Byte)
Dim SetByte As Byte
   SetByte = (pSeconds / 2) Or (pMinutes << 5)      // Create time & date - Seconds/Minutes
   InsertByteIntoBuffer(pSectorPos, SetByte)
   Inc(pSectorPos) 
   SetByte = (pMinutes >> 3) Or (pHours << 3)       // Create time & date - Minutes/Hours
   InsertByteIntoBuffer(pSectorPos, SetByte)
   Inc(pSectorPos) 
   SetByte = pDay Or (pMonth << 5)                  // Create time & date - Day/Month
   InsertByteIntoBuffer(pSectorPos, SetByte)
   Inc(pSectorPos) 
   SetByte = (pMonth >> 3) Or ((pYear + 20) << 1)   // Create time & date - Month/Year
   InsertByteIntoBuffer(pSectorPos, SetByte)
End Sub     
{
********************************************************************************
* Name    : SetFileTimeCreate (PRIVATE)                                        *
* Purpose : Set create time & date for file                                    *
********************************************************************************
}   
Sub SetFileTimeCreate(pSectorPos As Word, pDay, pMonth, pYear,
                      pHours, pMinutes, pSeconds, pHSeconds As Byte)
Dim SetByte As Byte
   SetByte = pHSeconds + ((pSeconds Mod 2) * 100)   // Create time & date - Milliseconds/Minutes
   InsertByteIntoBuffer(pSectorPos, SetByte)
   Inc(pSectorPos) 
   SetFileTimeModified(pSectorPos, pDay, pMonth, pYear, pHours, pMinutes, pSeconds)       
End Sub
{
********************************************************************************
* Name    : CloseFile (PUBLIC)                                                 *
* Purpose : Close file                                                         *
********************************************************************************
}   
Public Sub CloseFile(pDay As Byte = 1, pMonth As Byte = 1, pYear As Byte = 1,             
                     pHours As Byte = 0, pMinutes As Byte = 0, pSeconds As Byte = 0)
Dim IsDirectory As Byte
   If File(1).RW = 1 Then
   ElseIf File(1).IsOpen Then                   // File must be open for write
      If File(1).EOF Then                       // If at eof, need to fill rest of sector with zeros,
                                                // then will automatically write sector at end
         While File(1).CurrentSectorPos < $200 
            WriteSectorByte($00)
         Wend
      Else
         WriteSector(File(1).CurrentSector)
      EndIf
      ReadSector(File(1).FileEntryDirectorySector)
      IsDirectory = Shared.CurrentSectorBuffer(File(1).FileEntryDirectorySectorPos + $0B)
      IsDirectory = IsDirectory And %00010000   // Check if it is a directory being closed
      InsertByteIntoBuffer(File(1).FileEntryDirectorySectorPos + $1C, File(1).Size.Byte0)   // File size
      InsertByteIntoBuffer(File(1).FileEntryDirectorySectorPos + $1D, File(1).Size.Byte1)
      InsertByteIntoBuffer(File(1).FileEntryDirectorySectorPos + $1E, File(1).Size.Byte2)
      InsertByteIntoBuffer(File(1).FileEntryDirectorySectorPos + $1F, File(1).Size.Byte3)
      If (File(1).Size > 0) Or (IsDirectory = %00010000) Then
         InsertByteIntoBuffer(File(1).FileEntryDirectorySectorPos + $1A, File(1).FirstCluster.Byte0)   // Cluster number      
         InsertByteIntoBuffer(File(1).FileEntryDirectorySectorPos + $1B, File(1).FirstCluster.Byte1)    
         InsertByteIntoBuffer(File(1).FileEntryDirectorySectorPos + $14, File(1).FirstCluster.Byte2)   // High byte for FAT32 
         InsertByteIntoBuffer(File(1).FileEntryDirectorySectorPos + $15, File(1).FirstCluster.Byte3)    
      EndIf
      SetFileTimeModified(File(1).FileEntryDirectorySectorPos + $16, pDay, pMonth, pYear,
                          pHours, pMinutes, pSeconds)
                                                                                      // Set file modified
                                                                                      // time & date
      WriteSector(File(1).FileEntryDirectorySector)      
      If File(1).EOF Then
         If (File(1).Size > 0) Or (IsDirectory = %00010000) Then
            ReadSector(Disk.FAT1 + File(1).FATSector)
            AssembleFATSector(File(1).FATSector, File(1).FATSectorPos, File(1).ClusterSequenceIndex, 1)
            WriteSector(Disk.FAT1 + File(1).FATSector)                                // Write FAT1
            WriteSector(Disk.FAT2 + File(1).FATSector)                                // Write FAT2
         EndIf
      EndIf
   EndIf
   File(1).Size = 0
   File(1).IsOpen = False
End Sub        
{
********************************************************************************
* Name    : Cmd5541 (PRIVATE)                                                  *
* Purpose : Sends Cmd 55 followed by 41                                        *
********************************************************************************
}   
Function Cmd5541(pHCSBitSet As Boolean) As Byte
   Cmd5541 = SendCmd(CMD55, 0)
   If Cmd5541 > 1 Then
      Exit
   Else
      If pHCSBitSet Then
         Cmd5541 = SendCmd(CMD41, $40000000)
      Else
         Cmd5541 = SendCmd(CMD41, $00000000)
      EndIf
   EndIf
End Function
{
********************************************************************************
* Name    : InitSequence (PRIVATE)                                             *
* Purpose : Init SD or MMC card.                                               *
*         : Sets Input/Output pins and switches card to SPI mode.              *
*         : Return: 0 = Success, 1 = Fail/card not present.                    *
*         : Either software or hardware SPI can be used by selecting           *
*         : #option SD_SPI = SW or #option SD_SPI = MSSP                       *
********************************************************************************
}   
Function InitSequence() As Boolean
Dim Index As Word
Dim OCR(4) As Byte
   #if SD_SPI = MSSP                             // MSSP version
      SSPControl1 = %00000000                    // Turn off MSSP module for SW slow init
   #endif
   Disk.SlowSPI = 1                              // Set SPI to slow mode for initialisation
   Output(CS)                                    // Setup direction for SPI bus
   Output(DI)
   Output(CLK)
   Input(DO)
   InitSequence = False
   Disk.CardType = 0
   CS = 1                                        // Pull CS high
   Index = 10
   Repeat                                        // Clock SD for min 80 cycles
      SendByte($FF)
      Dec(Index)
   Until Index = 0
   DelayMS(100)
   CS = 0                                                // Pull CS low
   If SendCmd(CMD0, $00000000) = $01 Then                // Send Cmd 0 - enter idle state
      If SendCmd(CMD8, $000001AA) = $01 Then             // SDC Version 2+
         For Index = 0 To 3
            OCR(Index) = ReceiveByte()
         Next
         If OCR(2) = $01 And OCR(3) = $AA Then           // SD Vdd range of 2.7 - 3.6 volts
            Index = TimeOutLimit
            Repeat
               Dec(Index)               
            Until Cmd5541(True) = 0 Or Index = 0         
            If Index > 0 Then                            // Cmd 55/41 successful with HCS bit set
               If SendCmd(CMD58, $00000000) = 0 Then     // Send Cmd 58
                  For Index = 0 To 3
                     OCR(Index) = ReceiveByte()
                  Next
                  If (OCR(0) And $40) > 0 Then           // CCS[30] bit in the OCR set
                     Disk.CardType = sdSDHC              // Init successful - SDHC (3)
                  Else                 
                     Disk.CardType = sdSD                // Init successful - SD (2)
                  EndIf
               EndIf
            EndIf
         EndIf
      Else                                               // SD Version 1 or MMC
         If Cmd5541(False) <= 1 Then                     
            Disk.CardType = sdSD                         // Init successful - SD (2)
            Index = TimeOutLimit
            Repeat
               Dec(Index)               
            Until Cmd5541(True) = 0 Or Index = 0         // Wait for leaving idle state
         Else                  
            Disk.CardType = sdMMC                        // Init successful - MMC (1)
            Index = TimeOutLimit
            Repeat
               Dec(Index)               
            Until SendCmd(CMD1, $00) = 0 Or Index = 0    // Wait for leaving idle state
         EndIf
         If Index = 0 Then
            Disk.CardType = 0
         EndIf    
      EndIf
   EndIf
   CS = 1
   SendByte($FF)                                         // Clock SD/MMC to complete
   If Disk.CardType > 0 Then
      InitSequence = True
      #if SD_SPI = MSSP                                  // MSSP version
         SSPControl1 = SPISpeed Or %00100000             // Setup MSSP module & turn on
         SSPStatus = %01000000                         
      #endif
      Disk.SlowSPI = 0                                   // Set SPI to fast mode
   EndIf
End Function
{
********************************************************************************
* Name    : Init (PUBLIC)                                                      *
* Purpose : Init SD or MMC card and read FAT16/32 file system.                 *
*         : Sets Input/Output pins and switches card to SPI mode, then reads   *
*         : parameters of Master Boot Record & Boot Record.                    *
*         : Return: errOK = Success, errNoResponse = No response from card,    *
*         : errInvalidFormat = Card not formatted with FAT16 or FAT32,         *
*         : errRWError = Read/write error                                      *
*         : Either software or hardware SPI can be used by selecting           *
*         : #option SD_SPI = SW or #option SD_SPI = MSSP                       *
********************************************************************************
}   
Public Function Init() As Byte
Dim BootRecord As LongWord
Dim BytesPerSector As Word
Dim ReservedSectors As Word
Dim NumberOfFATs As Byte
Dim NumberOfRootEntries As Word
Dim NumberOfSectors As LongWord
Dim CountOfClusters As LongWord
Dim PartitionOffset As Word
Dim PartitionType As Byte
Dim PartitionFound As Boolean
Dim Index As Byte
   Disk.RWError = False                                            // Clear read/write error flag to allow a re-initialise after failure
   If Not InitSequence() Then                                      // Init SD/MMC
      Disk.RWError = True
      Result = errNoResponse
      Exit
   EndIf
   SectorInitRead($00)                                             // Read Master Boot Record               
   BootRecord = $0000                                           
   If Shared.CurrentSectorBuffer(0) = $EB And Shared.CurrentSectorBuffer(2) = $90 Then
                                                                   // Locate Boot Record - Sector 0 is Boot Record (no MBR)
   ElseIf Shared.CurrentSectorBuffer(0) = $E9 Then
                                                                   // Locate Boot Record - Sector 0 is Boot Record (no MBR)
   Else
      Index = 0
      PartitionFound = False
      Repeat
         PartitionOffset = $1C2 + (Index * 16)
         PartitionType = Shared.CurrentSectorBuffer(PartitionOffset)
         If (PartitionType = 4) Or (PartitionType = 6) Or (PartitionType = 14) Or (PartitionType = 11) Or (PartitionType = 12) Then   // Check valid FAT16 or FAT32 partition
            PartitionFound = True
         EndIf
         Inc(Index)
      Until (PartitionFound = True) Or (Index = 4)
      If PartitionFound Then
         BootRecord.Byte0 = Shared.CurrentSectorBuffer(PartitionOffset + 4)  // Locate Boot Record 
         BootRecord.Byte1 = Shared.CurrentSectorBuffer(PartitionOffset + 5)
         BootRecord.Byte2 = Shared.CurrentSectorBuffer(PartitionOffset + 6)
         BootRecord.Byte3 = Shared.CurrentSectorBuffer(PartitionOffset + 7)
      Else
         Disk.RWError = True
         Result = errInvalidFormat
         Exit
      EndIf
   EndIf                                                                                
   SectorInitRead(BootRecord)
   BytesPerSector.Byte0 = Shared.CurrentSectorBuffer($0B)         // Read bytes per sector           
   BytesPerSector.Byte1 = Shared.CurrentSectorBuffer($0C)
   If Not((BytesPerSector = 512) And ((Shared.CurrentSectorBuffer(0) = $EB And Shared.CurrentSectorBuffer(2) = $90) Or (Shared.CurrentSectorBuffer(0) = $E9)) And (Shared.CurrentSectorBuffer(510) = $55 And Shared.CurrentSectorBuffer(511) = $AA)) Then
      Disk.RWError = True
      Result = errInvalidFormat
      Exit
   EndIf                               
   Disk.SectorsPerCluster = Shared.CurrentSectorBuffer($0D)       // Read sectors per cluster                
   ReservedSectors.Byte0 = Shared.CurrentSectorBuffer($0E)        // Read number of reserved sectors
   ReservedSectors.Byte1 = Shared.CurrentSectorBuffer($0F)                                          
   NumberOfFATs = Shared.CurrentSectorBuffer($10)                 // Read number of FATs           
   NumberOfRootEntries.Byte0 = Shared.CurrentSectorBuffer($11)    // Read number of Root Entries      
   NumberOfRootEntries.Byte1 = Shared.CurrentSectorBuffer($12)
   Disk.SectorsPerFAT = 0      
   Disk.SectorsPerFAT.Byte0 = Shared.CurrentSectorBuffer($16)     // Read sectors Per FAT - FAT16           
   Disk.SectorsPerFAT.Byte1 = Shared.CurrentSectorBuffer($17)
   If Disk.SectorsPerFAT = 0 Then
      Disk.SectorsPerFAT.Byte0 = Shared.CurrentSectorBuffer($24)  // Read sectors Per FAT - FAT32           
      Disk.SectorsPerFAT.Byte1 = Shared.CurrentSectorBuffer($25)
      Disk.SectorsPerFAT.Byte2 = Shared.CurrentSectorBuffer($26)            
      Disk.SectorsPerFAT.Byte3 = Shared.CurrentSectorBuffer($27)
   EndIf                                
   NumberOfSectors = 0
   NumberOfSectors.Byte0 = Shared.CurrentSectorBuffer($13)        // Read number of sectors - 16 bit version                   
   NumberOfSectors.Byte1 = Shared.CurrentSectorBuffer($14)
   If NumberOfSectors = 0 Then                      
      NumberOfSectors.Byte0 = Shared.CurrentSectorBuffer($20)     // Read number of sectors - 32 bit version                    
      NumberOfSectors.Byte1 = Shared.CurrentSectorBuffer($21)                      
      NumberOfSectors.Byte2 = Shared.CurrentSectorBuffer($22)                      
      NumberOfSectors.Byte3 = Shared.CurrentSectorBuffer($23)
   EndIf
   Disk.FAT1 = BootRecord + ReservedSectors                        // Locate FAT1
   Disk.FAT2 = BootRecord + ReservedSectors + Disk.SectorsPerFAT   // Locate FAT2
   Disk.RootDirectory = NumberOfFATs * Disk.SectorsPerFAT          // Locate Root Directory
   Disk.RootDirectory = Disk.RootDirectory + BootRecord + ReservedSectors  
   Disk.SectorsInRoot = NumberOfRootEntries * 32                   // Calculate sectors in Root Directory
   Disk.SectorsInRoot = Disk.SectorsInRoot / BytesPerSector
   Disk.Data = Disk.RootDirectory + Disk.SectorsInRoot             // Locate start of data @ cluster 2
   File(1).CurrentSector = BootRecord + NumberOfSectors            // Locate last sector (File.CurrentSector = BootRecord + NumberOfSectors - 1) & add 1 to count full clusters correctly with DataToCluster
   Disk.LastCluster = DataToCluster() - 1                          // Subtract 1 to correct result from DataToCluster when used to find number of complete clusters
   File(1).Number = 0                                              // Set Dir to read first file on first call
   File(1).DirFileDir = 0                                          // Set Dir to look for file, not directory, by default 
   File(1).IsOpen = False
   Disk.IsDirty = False
   CountOfClusters = Disk.LastCluster - 1                          // Determine FAT type
   If CountOfClusters < 4085 Then
      Disk.FATType = 0
      Disk.RWError = True
      Result = errInvalidFormat
      Exit                                                         // Exclude FAT12
   ElseIf CountOfClusters < 65525 Then
      Disk.FATType = sdFAT16                                       // Allow FAT16
      File(1).IsRootDirectory = True
   Else
      Disk.FATType = sdFAT32                                       // Exclude FAT32
      Disk.RootDirectory.Byte0 = Shared.CurrentSectorBuffer($2C)   // Read Root Cluster for FAT32
      Disk.RootDirectory.Byte1 = Shared.CurrentSectorBuffer($2D)
      Disk.RootDirectory.Byte2 = Shared.CurrentSectorBuffer($2E)
      Disk.RootDirectory.Byte3 = Shared.CurrentSectorBuffer($2F)
      Disk.RootDirectory = ClusterToData(Disk.RootDirectory)       // Convert to first sector of Root Cluster
      File(1).IsRootDirectory = False                              // For FAT32, start off as if not in root directory; code will now follow cluster chain as for folders with FAT16
   EndIf
   File(1).DirectoryStartSector = Disk.RootDirectory               // Start off in Root Directory
   If Disk.RWError Then
      Result = errRWError
   Else
      Result = errOK
   EndIf
End Function
{
********************************************************************************
* Name    : ChDir (OVERLOAD)                                                   *
* Purpose : Change to directory in current directory.                          *
*         : Return: True = Success, False = Directory not found                *
********************************************************************************
}   
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then                                      
Public Function ChDir(pDirName As TSDName) As Boolean
   If Not(DirExists(pDirName)) Then             // Check if directory exists
      ChDir = False                             // Error - directory does not exist
      Exit                                                                    
   EndIf
   File(1).DirectoryStartSector = File(1).CurrentSector
   If (File(1).DirectoryStartSector = Disk.RootDirectory) And (Disk.FATType = 1) Then   // Only set IsRootDirectory = True for FAT16, keep as false for FAT32, even when in root cluster
      File(1).IsRootDirectory = True
   Else
      File(1).IsRootDirectory = False
   EndIf
   File(1).Number = 0                           // Reset file pointer for Dir
   ChDir = True                                 // Success - directory swapped
End Function                                                                        
#endif
{
********************************************************************************
* Name    : ChDir (OVERLOAD)                                                   *
* Purpose : Change to directory in current directory.                          *
*         : Return: True = Success, False = Directory not found                *
********************************************************************************
}   
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then                                      
Public Function ChDir(pDirName As String) As Boolean
   Result = ChDir(StrToFileDirName(pDirName))
End Function
#endif
{
********************************************************************************
* Name    : ChDir (OVERLOAD)                                                   *
* Purpose : Change to parent or root directory.                                *
*         : pCmd: cdUp = parent directory, cdRoot = Root Dir                   *
*         : Return: True = Success, False = Directory not found                *
********************************************************************************
}   
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then                                      
Public Function ChDir(pCmd As Byte) As Boolean
Dim TempName As TSDName
   If pCmd = cdUp Then
      TempName.Name = ".."
      Result = ChDir(TempName)
   Else
      File(1).DirectoryStartSector = Disk.RootDirectory
      If Disk.FATType = 1 Then
         File(1).IsRootDirectory = True
      EndIf
      File(1).Number = 0   // Reset file pointer for Dir
      Result = True
   EndIf
End Function
#endif
{
********************************************************************************
* Name    : IsRootDir (PUBLIC)                                                 *
* Purpose : Determines if current directory is Root Directory.                 *
*         : Return: False = Not Root Directory, True = Is Root Directory       *
********************************************************************************
}   
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then                                      
Public Function IsRootDir() As Boolean
   If File(1).DirectoryStartSector = Disk.RootDirectory Then
      Result = True
   Else
      Result = False
   EndIf
End Function   
#endif
{
********************************************************************************
* Name    : InsertFileNameIntoBuffer (PRIVATE)                                 *
* Purpose : Inserts a filename into the directory entry in the buffer          *
********************************************************************************
}   
Sub InsertFileNameIntoBuffer(ByRef pFileName As TSDName)
Dim RootDirFilePos As Byte
   File(1).CurrentSectorPos = File(1).FileEntryDirectorySectorPos
   RootDirFilePos = $00
   Repeat                                                                    // Write File Name
      InsertByteIntoBuffer(File(1).CurrentSectorPos, pFileName.Name(RootDirFilePos))      
      Inc(File(1).CurrentSectorPos)
      Inc(RootDirFilePos)
   Until RootDirFilePos > $07
   Repeat                                                                    // Write File Extension
      InsertByteIntoBuffer(File(1).CurrentSectorPos, pFileName.Extension(RootDirFilePos - $08))      
      Inc(File(1).CurrentSectorPos)
      Inc(RootDirFilePos)
   Until RootDirFilePos > $0A
End Sub
{
********************************************************************************
* Name    : New (PRIVATE)                                                      *
* Purpose : Create a new file or directory.                                    *
*         : pFileDir: 0 = file, 1 = directory                                  * 
*         : Return: errOK = Success, errExists = File already exists,          *
*         : errDiskFull = Disk full, errRootDirFull = Root Directory full,     *
*         : errInUse = File number in use
********************************************************************************
}   
Function New(pFileName As TSDName, pIsFile As Boolean,                          
             pDay As Byte = 1, pMonth As Byte = 1, pYear As Byte = 1,
             pHours As Byte = 0, pMinutes As Byte = 0, pSeconds As Byte = 0, pHSeconds As Byte = 0) As Byte
Dim Response As Byte
Dim ResponseB As Boolean
Dim RootDirFilePos As Byte
Dim Cluster As LongWord
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then
Dim NewDirSector As LongWord
Dim Index As Byte
#endif
   If Disk.RWError Then
      New = errRWError
      Exit
   EndIf
   If File(1).IsOpen Then
      New = errInUse
      Exit
   EndIf
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then
   If pIsFile Then
      ResponseB = FileExists(pFileName)                                   // Check if file exists
   Else
      ResponseB = DirExists(pFileName)                                    // Check if directory exists
   EndIf
#else
   ResponseB = FileExists(pFileName)                                      // Check if file exists
#endif
   If ResponseB Then
      New = errExists                                                     // Error - file already exists
      Exit                                                                    
   EndIf
   Repeat
      File(1).FATSector = 0
      File(1).FATSectorPos = 0
      File(1).ClusterSequenceNumber = 0
      Response = FindFreeCluster()
      If Response = 1 Then                                                // Disk full error
         New = errDiskFull
         Exit
      EndIf   
      Cluster = FATSectorToCluster()
      If Not(FindRootDirEntry(pFileName, 0)) Then                         // No free directory entry
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then
         If File(1).IsRootDirectory Then                                  // Is Root Directory - no 
                                                                          // possibility of expansion
            New = errRootDirFull
            Exit
         Else                                                             // Expand directory with free cluster already found
            WriteFATEntry(Cluster, File(1).FATSectorPos)                  // Point end of existing cluster chain to new cluster FAT sector set & read by IncrementSector in FindRootDirEntry
            WriteSector(Disk.FAT1 + File(1).FATSector)                    // Write to FAT1
            WriteSector(Disk.FAT2 + File(1).FATSector)                    // Write to FAT2
            ClusterToFATSector(Cluster)                                   // Find position of new cluster entry
            ReadSector(Disk.FAT1 + File(1).FATSector)
            WriteFATEntry($0FFFFFFF, File(1).FATSectorPos)                // End cluster chain
            WriteSector(Disk.FAT1 + File(1).FATSector)                    // Write to FAT1
            WriteSector(Disk.FAT2 + File(1).FATSector)                    // Write to FAT2
            Clear(Shared.CurrentSectorBuffer)
            NewDirSector = ClusterToData(Cluster)
            Index = 0
            Repeat
               WriteSector(NewDirSector + Index)
               Inc(Index)
            Until Index = Disk.SectorsPerCluster
         EndIf
#else
         New = errRootDirFull
         Exit
#endif
      Else
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then
         ClusterToFATSector(Cluster)                                  // Reset FAT sector parameters to file parameters, not directory parameters
#endif                                                               
      EndIf 
   Until Response = 0                                                 // Repeat finding free cluster
                                                                      // & directory entry if dir expanded
   ReadSector(File(1).FileEntryDirectorySector)
   InsertFileNameIntoBuffer(pFileName)
   File(1).CurrentSectorPos = File(1).FileEntryDirectorySectorPos + $0B
   RootDirFilePos = $0B
   Repeat                                                             // Erase Root Directory position
      InsertByteIntoBuffer(File(1).CurrentSectorPos, 0)      
      Inc(File(1).CurrentSectorPos)
      Inc(RootDirFilePos)
   Until RootDirFilePos > $1F
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then
   If pIsFile Then
      InsertByteIntoBuffer(File(1).FileEntryDirectorySectorPos + $0B, %00100000)
                                                                      // File attributes (00ADVSHR)
                                                                      // 0: unused bit, A: archive bit,
                                                                      // D: Dir bit, V: volume bit, 
                                                                      // S: system bit, R: read-only bit    
   Else
      InsertByteIntoBuffer(File(1).FileEntryDirectorySectorPos + $0B, %00010000) 
                                                                      // File attributes (00ADVSHR)
                                                                      // 0: unused bit, A: archive bit, 
                                                                      // D: Dir bit, V: volume bit, 
                                                                      // S: system bit, R: read-only bit    
   EndIf
#else
   InsertByteIntoBuffer(File(1).FileEntryDirectorySectorPos + $0B, %00100000)
                                                                      // File attributes (00ADVSHR)
                                                                      // 0: unused bit, A: archive bit, 
                                                                      // D: Dir bit, V: volume bit,
                                                                      // S: system bit, R: read-only bit    
#endif
   SetFileTimeCreate(File(1).FileEntryDirectorySectorPos + $0D, pDay, pMonth, pYear,
                     pHours, pMinutes, pSeconds, pHSeconds)
                                                                      // Set file create time & date
   File(1).FirstCluster = Cluster                                     // Store first cluster
   WriteSector(File(1).FileEntryDirectorySector)                      
   File(1).CurrentSector = ClusterToData(Cluster)                     // Set start sector for data
   SectorInitWrite()
   File(1).CurrentSectorInCluster = 1
   File(1).CurrentSectorPos = 0                                       // Set start pos for sector buffer
   File(1).Size = 0
   File(1).BytesRead = 0
   File(1).EOF = True
   Disk.DiskFull = False
   File(1).IsOpen = True
   File(1).RW = 0
   New = errOK
End Function
{
********************************************************************************
* Name    : NewFile (OVERLOAD)                                                 *
* Purpose : Create a new file.                                                 *
*         : Return: errOK = Success, errExists = File already exists,          *
*         : errDiskFull = Disk full, errRootDirFull = Root Directory full,     *
*         : errInUse = File number in use, errRWError = Read/write error       *
********************************************************************************
}   
Public Function NewFile(pFileName As TSDName, pDay As Byte = 1, pMonth As Byte = 1, pYear As Byte = 1,     
                        pHours As Byte = 0, pMinutes As Byte = 0, pSeconds As Byte = 0,
                        pHSeconds As Byte = 0) As Byte
   Result = New(pFileName, True, pDay, pMonth, pYear, pHours, pMinutes, pSeconds, pHSeconds)
End Function
{
********************************************************************************
* Name    : NewFile (OVERLOAD)                                                 *
* Purpose : Create a new file.                                                 *
*         : Return: errOK = Success, errExists = File already exists,          *
*         : errDiskFull = Disk full, errRootDirFull = Root Directory full,     *
*         : errInUse = File Number in use, errRWError = Read/Write error       *
********************************************************************************
}   
Public Function NewFile(pFileName As String, pDay As Byte = 1, pMonth As Byte = 1, pYear As Byte = 1,      
                        pHours As Byte = 0, pMinutes As Byte = 0, pSeconds As Byte = 0, 
                        pHSeconds As Byte = 0) As Byte
   Result = New(StrToFileDirName(pFileName), True, pDay, pMonth, pYear,
                pHours, pMinutes, pSeconds, pHSeconds)
End Function
{
********************************************************************************
* Name    : MkDir (OVERLOAD)                                                   *
* Purpose : Make a new directory.                                              *
*         : Return: errOK = Success, errExists = Directory already exists,     *
*         : errDiskFull = Disk full, errRootDirFull = Root Directory full,     *
*         : errInUse = File number in use, errRWError = Read/write error       *
********************************************************************************
}   
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then                                      
Public Function MkDir(pDirName As TSDName, pDay As Byte = 1, pMonth As Byte = 1, pYear As Byte = 1,
                      pHours As Byte = 0, pMinutes As Byte = 0, pSeconds As Byte = 0, 
                      pHSeconds As Byte = 0) As Byte
Dim Index As Byte
Dim Cluster As LongWord
Dim TempCurrentSector As LongWord
   pDirName.Extension = ""
   MkDir = New(pDirName, False, pDay, pMonth, pYear, pHours, pMinutes, pSeconds, pHSeconds)
   If MkDir > 0 Then
      Exit
   EndIf
   Clear(Shared.CurrentSectorBuffer)
   Index = $00
   Repeat
      InsertByteIntoBuffer(Index, $20)   // Write 11 spaces to clear first directory entries (for ".")
      Inc(Index)
   Until Index > $0A
   Index = $20
   Repeat
      InsertByteIntoBuffer(Index, $20)   // Write 11 spaces to clear second directory entries (for "..")
      Inc(Index)
   Until Index > $2A
   InsertByteIntoBuffer($00,$2E)   // "."
   InsertByteIntoBuffer($20,$2E)   // ".."
   InsertByteIntoBuffer($21,$2E)
   InsertByteIntoBuffer($0B,$10)   // Dir attribute for "."
   InsertByteIntoBuffer($2B,$10)   // Dir attribute for ".."
   SetFileTimeCreate($0D, pDay, pMonth, pYear, pHours, pMinutes, pSeconds, pHSeconds)   // Set file create
                                                                                        // time & date
                                                                                        // for "."
   SetFileTimeCreate($2D, pDay, pMonth, pYear, pHours, pMinutes, pSeconds, pHSeconds)   // Set file create
                                                                                        // time & date
                                                                                        // for ".."
   SetFileTimeModified($16, pDay, pMonth, pYear, pHours, pMinutes, pSeconds)   // Set file modified time
                                                                               // & date for "."
   SetFileTimeModified($36, pDay, pMonth, pYear, pHours, pMinutes, pSeconds)   // Set file modified time
                                                                               // & date for ".."
   Cluster = DataToCluster()
   InsertByteIntoBuffer($1A, Cluster.Byte0)   // Cluster number for "."      
   InsertByteIntoBuffer($1B, Cluster.Byte1)    
   InsertByteIntoBuffer($14, Cluster.Byte2)   // High byte for FAT32
   InsertByteIntoBuffer($15, Cluster.Byte3)    
   TempCurrentSector = File(1).CurrentSector
   If (File(1).DirectoryStartSector = Disk.RootDirectory) Then   // Set to point to cluster 0 if parent is root directory if using FAT16 or FAT32
      Cluster = 0
   Else
      File(1).CurrentSector = File(1).DirectoryStartSector
      Cluster = DataToCluster()
      File(1).CurrentSector = TempCurrentSector     
   EndIf
   InsertByteIntoBuffer($3A, Cluster.Byte0)   // Cluster number for ".."     
   InsertByteIntoBuffer($3B, Cluster.Byte1)    
   InsertByteIntoBuffer($34, Cluster.Byte2)   // High byte for FAT32     
   InsertByteIntoBuffer($35, Cluster.Byte3)    
   File(1).CurrentSectorPos = $1FF            // Leave CurrentSectorPos as $1FF to force sector write
   CloseFile()
   Clear(Shared.CurrentSectorBuffer)
   Index = 1
   Repeat                                     // Clear remaining sectors of first cluster in directory
      WriteSector(TempCurrentSector + Index)
      Inc(Index)
   Until Index = Disk.SectorsPerCluster 
End Function
#endif
{
********************************************************************************
* Name    : MkDir (OVERLOAD)                                                   *
* Purpose : Make a new directory.                                              *
*         : Return: errOK = Success, errExists = Directory already exists,     *
*         : errDiskFull = Disk full, errRootDirFull = Root Directory full,     *
*         : errInUse = File Number in use, errRWError = Read/Write error       *
********************************************************************************
}   
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then                                      
Public Function MkDir(pDirName As String, pDay As Byte = 1, pMonth As Byte = 1, pYear As Byte = 1,
                      pHours As Byte = 0, pMinutes As Byte = 0, pSeconds As Byte = 0,
                      pHSeconds As Byte = 0) As Byte
   Result = MkDir(StrToFileDirName(pDirName), pDay, pMonth, pYear, pHours, pMinutes, pSeconds, pHSeconds)
End Function
#endif
{
********************************************************************************
* Name    : OpenFile (OVERLOAD)                                                *
* Purpose : Open an existing file.                                             *
*         : Return: errOK = Success, errNotFound = File not found,             *
*         : errInUse = File number in use, errRWError = Read/write error       *
********************************************************************************
}   
Public Function OpenFile(pFileName As TSDName) As Byte                          
   If Disk.RWError Then
      OpenFile = errRWError
      Exit
   EndIf
   If File(1).IsOpen Then
      OpenFile = errInUse
      Exit
   EndIf
   If Not(FileExists(pFileName)) Then           // Check if file exists
      OpenFile = errNotFound                    // Error - file does not exist
      Exit                                                                    
   EndIf
   SectorInitRead(File(1).CurrentSector)
   File(1).CurrentSectorInCluster = 1
   File(1).CurrentSectorPos = 0                 // Set start position for sector 
   File(1).BytesRead = 0                        // Zero sequential read marker
   If File(1).Size > 0 Then
      File(1).EOF = False
   Else
      File(1).EOF = True
   EndIf
   File(1).IsOpen = True
   File(1).RW = 1
   OpenFile = errOK                             // Success - file open for reading
End Function                                                                        
{
********************************************************************************
* Name    : OpenFile (OVERLOAD)                                                *
* Purpose : Open an existing file.                                             *
*         : Return: errOK = Success, errNotFound = File Not found,             *
*         : errInUse = File Number in use, errRWError = Read/Write error       *
********************************************************************************
}   
Public Function OpenFile(pFileName As String) As Byte                           
   Result = OpenFile(StrToFileDirName(pFileName))
End Function
{
********************************************************************************
* Name    : ReadByte (PUBLIC)                                                  *
* Purpose : Read a single byte from an existing file                           *
*         : File.EOF = True when EOF condition met                             *
********************************************************************************
}   
Public Function ReadByte() As Byte                                              
   If File(1).EOF Then
      ReadByte = 0
      Exit
   EndIf
   Inc(File(1).BytesRead)
   If File(1).BytesRead >= File(1).Size Then
      File(1).EOF = True
   EndIf
   If File(1).CurrentSectorPos > $1FF Then
      If Disk.IsDirty Then                      // If bytes have been written in current sector,
                                                // write sector back to SD card and reset Disk.IsDirty
         WriteSector(File(1).CurrentSector)
         Disk.IsDirty = False
      EndIf
      IncrementSector()
      SectorInitRead(File(1).CurrentSector)
   EndIf
   ReadByte = Shared.CurrentSectorBuffer(File(1).CurrentSectorPos)
   Inc(File(1).CurrentSectorPos)
End Function
{
********************************************************************************
* Name    : EOF (PUBLIC)                                                       *
* Purpose : Determines if a file opened for reading has reached the end of     *
*         : file (EOF), or if RWError to prevent blocking loop when reading.   *
*         : Return: False = EOF not yet reached, True = EOF reached            *
********************************************************************************
}   
Public Function EOF() As Boolean
   Result = File(1).EOF Or Disk.RWError
End Function   
{
********************************************************************************
* Name    : ReadBoolean (PUBLIC)                                               *
* Purpose : Read a Boolean from file.                                          *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Public Function ReadBoolean() As Boolean
   ReadBoolean = Boolean(ReadByte())
End Function
{
********************************************************************************
* Name    : ReadBit (PUBLIC)                                                   *
* Purpose : Read a Bit from file.                                              *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Public Function ReadBit() As Bit
   ReadBit = Bit(ReadByte())
End Function
{
********************************************************************************
* Name    : ReadChar (PUBLIC)                                                  *
* Purpose : Read a Char from file.                                             *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Public Function ReadChar() As Char
   ReadChar = ReadByte()
End Function
{
********************************************************************************
* Name    : ReadShortInt (PUBLIC)                                              *
* Purpose : Read a ShortInt from file.                                         *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Public Function ReadShortInt() As ShortInt
   ReadShortInt = ReadByte()
End Function
{
********************************************************************************
* Name    : ReadWord (PUBLIC)                                                  *
* Purpose : Read a Word from file.                                             *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Public Function ReadWord() As Word
   ReadWord.Byte0 = ReadByte()
   ReadWord.Byte1 = ReadByte()
End Function
{
********************************************************************************
* Name    : ReadInteger (PUBLIC)                                               *
* Purpose : Read a Integer from file.                                          *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Public Function ReadInteger() As Integer
   ReadInteger.Byte0 = ReadByte()
   ReadInteger.Byte1 = ReadByte()
End Function
{
********************************************************************************
* Name    : ReadLongWord (PUBLIC)                                              *
* Purpose : Read a LongWord from file.                                         *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Public Function ReadLongWord() As LongWord
   ReadLongWord.Byte0 = ReadByte()
   ReadLongWord.Byte1 = ReadByte()
   ReadLongWord.Byte2 = ReadByte()
   ReadLongWord.Byte3 = ReadByte()
End Function
{
********************************************************************************
* Name    : ReadLongInt (PUBLIC)                                               *
* Purpose : Read a LongInt from file.                                          *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Public Function ReadLongInt() As LongInt
   ReadLongInt.Byte0 = ReadByte()
   ReadLongInt.Byte1 = ReadByte()
   ReadLongInt.Byte2 = ReadByte()
   ReadLongInt.Byte3 = ReadByte()
End Function
{
********************************************************************************
* Name    : ReadFloat (PUBLIC)                                                 *
* Purpose : Read a Float from file.                                            *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Public Function ReadFloat() As Float
   ReadFloat.Byte0 = ReadByte()
   ReadFloat.Byte1 = ReadByte()
   ReadFloat.Byte2 = ReadByte()
   ReadFloat.Byte3 = ReadByte()
End Function
{
********************************************************************************
* Name    : ReadNBytes (PUBLIC)                                                *
* Purpose : Read N bytes from file.                                            *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Public Sub ReadNBytes(ByRef pReturnArray() As Byte, pNumberOfBytes As Word) 
Dim Address As Word
Dim FileByte As Byte
   Address = AddressOf(pReturnArray(0)) 
   While pNumberOfBytes > 0
      FileByte = SD.ReadByte()
      FSR0 = Address 
      POSTINC0 = FileByte 
      Dec(pNumberOfBytes)
      Inc(Address) 
   Wend 
End Sub 
{
********************************************************************************
* Name    : ReadItem (OVERLOAD)                                                *
* Purpose : Read a Boolean from file.                                          *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Sub ReadItem(ByRef pValue As Boolean)
   pValue = Boolean(ReadByte())
End Sub
{
********************************************************************************
* Name    : ReadItem (OVERLOAD)                                                *
* Purpose : Read a Bit from file.                                              *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Sub ReadItem(ByRef pValue As Bit)
   pValue = Bit(ReadByte())
End Sub
{
********************************************************************************
* Name    : ReadItem (OVERLOAD)                                                *
* Purpose : Read a Byte from file.                                             *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Sub ReadItem(ByRef pValue As Byte)
   pValue = ReadByte()
End Sub
{
********************************************************************************
* Name    : ReadItem (OVERLOAD)                                                *
* Purpose : Read a Char from file.                                             *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Sub ReadItem(ByRef pValue As Char)
   pValue = ReadByte()
End Sub
{
********************************************************************************
* Name    : ReadItem (OVERLOAD)                                                *
* Purpose : Read a ShortInt from file.                                         *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Sub ReadItem(ByRef pValue As ShortInt)
   pValue = ReadByte()
End Sub
{
********************************************************************************
* Name    : ReadItem (OVERLOAD)                                                *
* Purpose : Read a Word from file.                                             *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Sub ReadItem(ByRef pValue As Word)
   pValue = ReadWord()
End Sub
{
********************************************************************************
* Name    : ReadItem (OVERLOAD)                                                *
* Purpose : Read a Integer from file.                                          *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Sub ReadItem(ByRef pValue As Integer)
   pValue = ReadWord()
End Sub
{
********************************************************************************
* Name    : ReadItem (OVERLOAD)                                                *
* Purpose : Read a LongWord from file.                                         *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Sub ReadItem(ByRef pValue As LongWord)
   pValue = ReadLongWord()
End Sub
{
********************************************************************************
* Name    : ReadItem (OVERLOAD)                                                *
* Purpose : Read a LongInt from file.                                          *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Sub ReadItem(ByRef pValue As LongInt)
   pValue = ReadLongWord()
End Sub
{
********************************************************************************
* Name    : ReadItem (OVERLOAD)                                                *
* Purpose : Read a Float from file.                                            *
*         : Check EOF for EOF condition                                        *
********************************************************************************
}  
Sub ReadItem(ByRef pValue As Float)
   pValue = ReadFloat()
End Sub
{
********************************************************************************
* Name    : Read (COMPOUND)                                                    *
* Purpose : Read one or more items from file.                                  *
*         : Check EOF for EOF condition                                        *
********************************************************************************
} 
Public Compound Sub Read(ReadItem)
{
********************************************************************************
* Name    : Dir (OVERLOAD)                                                     *
* Purpose : Read file names from SD/MMC card.                                  *
*         : pDirDirection: dirFirst = First, dirNext = Next,                   *
*         : dirPrevious = Previous.                                            *
*         : pFileDir: sdFile = Find file, sdDirectory = Find directory.        *
*         : Return: If file found, returned as pFileName, if no file found,    *
*         : pFileName.Name = Null, pFileName.Extension = Null                  *
********************************************************************************
}   
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then                                      
Public Sub Dir(ByRef pName As TSDName, pDirDirection As Byte = dirNext, pFileDir As Bit = sdFile)
#else
Public Sub Dir(ByRef pName As TSDName, pDirDirection As Byte = dirNext)
#endif
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then
   If pFileDir <> File(1).DirFileDir Then
      File(1).Number = 0   // If swap from from searching for file to directory or vice versa,
                           // reset file number
      File(1).DirFileDir = pFileDir
   EndIf
#endif   
   Select pDirDirection
   Case 0
      File(1).Number = 1
   Case 1 
      Inc(File(1).Number)
   Case 2
      Dec(File(1).Number)
   EndSelect
   pName.Name = Null
   pName.Extension = Null
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then
   If Not(FindRootDirEntry(pName, 2, pFileDir)) Then
#else
   If Not(FindRootDirEntry(pName, 2, 0)) Then
#endif
      pName.Name = Null
      pName.Extension = Null
   EndIf
End Sub
{
********************************************************************************
* Name    : Dir (OVERLOAD)                                                     *
* Purpose : Read file names from SD/MMC card.                                  *
*         : pDirDirection: dirFirst = First, dirNext = Next,                   *
*         : dirPrevious = Previous.                                            *
*         : pFileDir: sdFile = Find File, sdDirectory = Find directory.        *
*         : Return: If file found, returned as string, e.g FILE0001.TXT,       *
*         : if no file found, return string is Null                            *
********************************************************************************
}   
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then                                      
Public Function Dir(pDirDirection As Byte = dirNext, pFileDir As Bit = sdFile) As String * 13
#else
Public Function Dir(pDirDirection As Byte = dirNext) As String * 13
#endif
Dim TempName As TSDName
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then
   Dir(TempName, pDirDirection, pFileDir)
#else
   Dir(TempName, pDirDirection)
#endif
   If TempName.Name <> Null Then
      Result = TempName.Name 
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then
      If pFileDir = sdFile Then
         Result = Result + "." + TempName.Extension
      EndIf
#else
      Result = Result + "." + TempName.Extension
#endif
   Else
      Result = Null
   EndIf
End Function
{
********************************************************************************
* Name    : TraverseClusters (PRIVATE)                                         *
* Purpose : Traverse cluster chain in FAT                                      *
********************************************************************************
}   
Sub TraverseClusters(pNumBytesToTraverse As LongWord)                           
Dim Cluster As LongWord
   Cluster = File(1).FirstCluster
   File(1).CurrentSector = 0
   While pNumBytesToTraverse > ($200 * Disk.SectorsPerCluster)
      ClusterToFATSector(Cluster)                                // Find FAT position for cluster
      If File(1).CurrentSector <> (Disk.FAT1 + File(1).FATSector) Then
         File(1).CurrentSector = Disk.FAT1 + File(1).FATSector
         SectorInitRead(File(1).CurrentSector)
      EndIf     
      Cluster = ReadFATEntry(File(1).FATSectorPos)
      pNumBytesToTraverse = pNumBytesToTraverse - ($200 * Disk.SectorsPerCluster)
   Wend
   ClusterToFATSector(Cluster)                                   // Find FAT position for last cluster
   If pNumBytesToTraverse = 0 Then                               // Test for zero length file - others
                                                                 // will not meet this
                                                                 // condition, as 1 or more bytes
                                                                 // will be left after while - wend loop
      File(1).CurrentSectorInCluster = 0
   Else
      File(1).CurrentSectorInCluster = (pNumBytesToTraverse - 1) / $200
   EndIf
   File(1).CurrentSector = ClusterToData(Cluster)                // Set append sector for data
   File(1).CurrentSector = File(1).CurrentSector + File(1).CurrentSectorInCluster 
   File(1).CurrentSectorInCluster = File(1).CurrentSectorInCluster + 1
   SectorInitRead(File(1).CurrentSector)
   File(1).CurrentSectorPos = pNumBytesToTraverse Mod $200       // Set append position for sector 
   If File(1).CurrentSectorPos = 0 Then                                              
      If pNumBytesToTraverse > 0 Then                            // Test not zero length file - others will 
                                                                 // meet this condition
         File(1).CurrentSectorPos = $200
         Inc(File(1).CurrentSector)
      EndIf
   EndIf
   File(1).ClusterSequenceIndex = 1
   File(1).ClusterSequenceNumber = 1
End Sub
{
********************************************************************************
* Name    : FSeek (PUBLIC)                                                     *
* Purpose : Seeks position in open file using pFilePtr.                        *
*         : Return: errOK = Success, errRWError = Read/Write error,            *
*         : errFileNotOpen = File not open, errBeyondEOF = Failed attemp to    *
*         : seek position beyond end of file                                   *
********************************************************************************
}   
Public Function FSeek(pFilePtr As LongWord) As Byte                             
Dim FileSizeHold As LongWord
   If Disk.RWError Then
      FSeek = errRWError
   ElseIf File(1).IsOpen = False Then
      FSeek = errFileNotOpen
   Else
      If Disk.IsDirty Then                                             // If bytes have been written in
                                                                       // current sector, write sector back
                                                                       // to SD card and reset IsDirty
         WriteSector(File(1).CurrentSector)
         Disk.IsDirty = False
      EndIf  
      If File(1).EOF Then                                              // Are at end or have extended file
         FileSizeHold = File(1).Size
         CloseFile()
         File(1).Size = FileSizeHold
         File(1).IsOpen = True
      EndIf
      If pFilePtr > File(1).Size Then
         FSeek = errBeyondEOF
      Else
         TraverseClusters(pFilePtr)
         File(1).BytesRead = pFilePtr
         If pFilePtr = File(1).Size Then
            File(1).EOF = True
         Else
            File(1).EOF = False
         EndIf
         FSeek = errOK                                                 // Success - file seek OK    
      EndIf
   EndIf
End Function
{
********************************************************************************
* Name    : OpenFileRW (OVERLOAD)                                              *
* Purpose : Open an existing file for reading & writing.                       *
*         : Return: errOK = Success, errNotFound = File not found,             *
*         : errDiskFull = Disk full, errInUse = File in use,                   *
*         : errRWError = Read/write error                                      *
********************************************************************************
}   
Public Function OpenFileRW(pFileName As TSDName) As Byte                        
Dim Response As Byte
Dim Cluster As LongWord
   If Disk.RWError Then
      OpenFileRW = errRWError
   ElseIf File(1).IsOpen Then
      OpenFileRW = errInUse
   ElseIf Not(FileExists(pFileName)) Then                              // Check if file exists & loads info
                                                                       // r.e. file
      OpenFileRW = errNotFound                                         // Error - file does not exist
   Else
      If File(1).Size = 0 Then
         File(1).FATSector = 0
         File(1).FATSectorPos = 0
         File(1).ClusterSequenceNumber = 0
         Response = FindFreeCluster()
         If Response = 1 Then                                          // Disk full error
            OpenFileRW = errDiskFull
            Exit
         EndIf   
         Cluster = FATSectorToCluster() 
         File(1).FirstCluster = Cluster                                // Store first cluster
         File(1).CurrentSector = ClusterToData(Cluster)                // Set start sector for data
         SectorInitWrite()
         File(1).CurrentSectorInCluster = 1
         File(1).CurrentSectorPos = 0
         File(1).EOF = True     
      Else
         File(1).FirstCluster = DataToCluster()
         TraverseClusters(0)                                           // Move to start of file
         File(1).EOF = False
      EndIf
      Disk.DiskFull = False
      File(1).IsOpen = True
      File(1).RW = 0
      File(1).BytesRead = 0
      Disk.IsDirty = False
      OpenFileRW = errOK                                               // Success - file open for appending
   EndIf
End Function
{
********************************************************************************
* Name    : OpenFileRW (OVERLOAD)                                              *
* Purpose : Open an existing file for reading & writing.                       *
*         : Return: errOK = Success, errNotFound = File not found,             *
*         : errDiskFull = Disk full, errInUse = File in use,                   *
*         : errRWError = Read/write error                                      *
********************************************************************************
}   
Public Function OpenFileRW(pFileName As String) As Byte
   Result = OpenFileRW(StrToFileDirName(pFileName))
End Function
{
********************************************************************************
* Name    : AppendFile (OVERLOAD)                                              *
* Purpose : Open an existing file for writing to end of file.                  *
*         : Return: errOK = Success, errNotFound = File not found,             *
*         : errDiskFull = Disk full, errInUse = File in use,                   *
*         : errRWError = Read/write error                                      *
********************************************************************************
}   
Public Function AppendFile(pFileName As TSDName) As Byte
   Result = OpenFileRW(pFileName)
   FSeek(File(1).Size)
End Function
{
********************************************************************************
* Name    : AppendFile (OVERLOAD)                                              *
* Purpose : Open an existing file for writing to end of file.                  *
*         : Return: errOK = Success, errNotFound = File Not found,             *
*         : errDiskFull = Disk full, errInUse = File in use,                   *
*         : errRWError = Read/Write error                                      *
********************************************************************************
}   
Public Function AppendFile(pFileName As String) As Byte
   Result = OpenFileRW(pFileName)
   FSeek(File(1).Size)
End Function
{
********************************************************************************
* Name    : DiskSizeKB (PUBLIC)                                                *
* Purpose : Find size of SD/MMC card (in KB)                                   *
********************************************************************************
}   
Public Function DiskSizeKB() As LongWord
   DiskSizeKB = (Disk.LastCluster - 1) * Disk.SectorsPerCluster
   DiskSizeKB = DiskSizeKB / 2
End Function
{
********************************************************************************
* Name    : FreeSpaceKB (PUBLIC)                                               *
* Purpose : Find free space on SD/MMC card (in KB)                             *
********************************************************************************
}   
Public Function FreeSpaceKB() As LongWord                                       
Dim FreeClusterNumber As LongWord
Dim ScanSector As LongWord
Dim ScanSectorPos As Word
Dim Correction As LongWord
   FreeClusterNumber = 0
   For ScanSector = Disk.FAT1 To (Disk.FAT1 + Disk.SectorsPerFAT - 1)   // Scan FAT1 for free clusters
      File(1).CurrentSector = ScanSector
      SectorInitRead(File(1).CurrentSector)
      ScanSectorPos = $000
      Repeat
         If ReadFATEntry(ScanSectorPos) = $00000000 Then
            Inc(FreeClusterNumber)
         EndIf
         If Disk.FATType = 1 Then
            Inc(ScanSectorPos, 2)   // FAT16
         Else
            Inc(ScanSectorPos, 4)   // FAT32
         EndIf
      Until ScanSectorPos > $1FF
   Next
   Correction = Disk.SectorsPerFAT * $100 / Disk.FATType                               
   Correction = Correction - Disk.LastCluster - 1                      
   FreeSpaceKB = FreeClusterNumber - Correction
   FreeSpaceKB = FreeSpaceKB * Disk.SectorsPerCluster
   FreeSpaceKB = FreeSpaceKB / 2
End Function
{
********************************************************************************
* Name    : Delete (PRIVATE)                                                   *
* Purpose : Delete an existing file or directory.                              *
*         : Return: True = Success, False = File not found                     *
********************************************************************************
}   
Function Delete(pFileName As TSDName, pIsFile As Boolean) As Boolean            
Dim ResponseB As Boolean
Dim FirstPass As Boolean
Dim Cluster As LongWord
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then
   If pIsFile Then
      ResponseB = FileExists(pFileName)                            // Check if file exists
   Else
      ResponseB = DirExists(pFileName)                             // Check if directory exists
   EndIf                                                           
#else
   ResponseB = FileExists(pFileName)                               // Check if file exists
#endif
   If Not(ResponseB) Then
      Delete = False                                               // Error - file/directory does not exist
      Exit                                                                    
   EndIf
   InsertByteIntoBuffer(File(1).FileEntryDirectorySectorPos, $E5)    
   WriteSector(File(1).FileEntryDirectorySector)
   If File(1).Size = 0 And pIsfile Then                            // Zero-length file - no need to
                                                                   // modify FAT
      Delete = True
      Exit
   EndIf
   Cluster = DataToCluster()
   FirstPass = True                                                // Use as flag for first pass
                                                                   // through repeat-until loop
   Repeat
      ClusterToFATSector(Cluster)                                  // Find next FAT position
      If File(1).CurrentSector <> (Disk.FAT1 + File(1).FATSector) Then   // Accessing new sector
         If FirstPass = False Then
            WriteSector(File(1).CurrentSector)                     // Write FAT1 sector before read next 
                                                                   // sector, except on first pass
            File(1).CurrentSector = File(1).CurrentSector - Disk.FAT1    // Copy FAT1 sector to FAT2
            File(1).CurrentSector = File(1).CurrentSector + Disk.FAT2
            WriteSector(File(1).CurrentSector)                     // Write FAT2 sector before read next 
                                                                   // sector, except on first pass
         EndIf
         File(1).CurrentSector = Disk.FAT1 + File(1).FATSector
         FirstPass = False
         SectorInitRead(File(1).CurrentSector)
      EndIf
      Cluster = ReadFATEntry(File(1).FATSectorPos)
      WriteFATEntry($00000000, File(1).FATSectorPos)               // Clear FAT entry
   Until (Disk.FATType = 1 And Cluster = $FFFF) Or (Disk.FATType = 2 And Cluster = $0FFFFFFF)
   WriteSector(File(1).CurrentSector)                              // Write last FAT1 sector
   File(1).CurrentSector = File(1).CurrentSector - Disk.FAT1       // Copy FAT1 sector to FAT2
   File(1).CurrentSector = File(1).CurrentSector + Disk.FAT2
   WriteSector(File(1).CurrentSector)                              // Write last FAT2 sector
   Delete = True                                                   // Success - file deleted
End Function
{
********************************************************************************
* Name    : DeleteFile (OVERLOAD)                                              *
* Purpose : Delete an existing file.                                           *
*         : Return: True = Success, False = File not found                     *
********************************************************************************
}   
Public Function DeleteFile(pFileName As TSDName) As Boolean
   Result = Delete(pFileName, True)
End Function
{
********************************************************************************
* Name    : DeleteFile (OVERLOAD)                                              *
* Purpose : Delete an existing file.                                           *
*         : Return: True = Success, False = File not found                     *
********************************************************************************
}   
Public Function DeleteFile(pFileName As String) As Boolean
   Result = Delete(StrToFileDirName(pFileName), True)
End Function
{
********************************************************************************
* Name    : QuickFormat (PUBLIC)                                               *
* Purpose : Quick format SD/MMC card.                                          *
*         : Will delete all files, but does not rebuild a corrupt file         *
*         : structure.                                                         *
*         : Return: True = Success, False = Read/write error                   *
********************************************************************************
}   
Public Function QuickFormat() As Boolean                                        
   QuickFormat = False
   If Disk.RWError Then
      Exit
   EndIf
   Clear(Shared.CurrentSectorBuffer)
   If Disk.FATType = 1 Then   // FAT16
      For File(1).CurrentSector = Disk.RootDirectory To (Disk.RootDirectory + Disk.SectorsInRoot - 1)   // Scan Root Directory & delete
         WriteSector(File(1).CurrentSector)
      Next
   Else                       // FAT32
      For File(1).CurrentSector = Disk.RootDirectory To (Disk.RootDirectory + Disk.SectorsPerCluster - 1)   // Scan first cluster for Root Directory folder & delete
         WriteSector(File(1).CurrentSector)
      Next   
   EndIf
   For File(1).CurrentSector = (Disk.FAT1 + 1) To (Disk.FAT1 + Disk.SectorsPerFAT - 1)  // Scan FAT1 & delete
      WriteSector(File(1).CurrentSector)
   Next
   For File(1).CurrentSector = (Disk.FAT2 + 1) To (Disk.FAT2 + Disk.SectorsPerFAT - 1)  // Scan FAT2 & delete
      WriteSector(File(1).CurrentSector)
   Next
   WriteFATEntry($0FFFFFF8, 0)
   WriteFATEntry($0FFFFFFF, 2 * Disk.FATType)
   WriteSector(Disk.FAT1)
   WriteSector(Disk.FAT2)
   If Disk.FATType = 2 Then   // FAT32 - re-write end of cluster chain for root directory folder
      File(1).CurrentSector = Disk.RootDirectory
      ClusterToFATSector(DataToCluster())
      ReadSector(Disk.FAT1 + File(1).FATSector)
      WriteFATEntry($0FFFFFFF, File(1).FATSectorPos)
      WriteSector(Disk.FAT1 + File(1).FATSector)
      WriteSector(Disk.FAT2 + File(1).FATSector)
   EndIf
   If Not Disk.RWError Then
      QuickFormat = True
   EndIf
End Function
{
********************************************************************************
* Name    : SaveFile(PUBLIC)                                                   *
* Purpose : Write all of buffer to file and update file size & FAT clusters.   *
*         : If disk removed, file should be safe. Has same effect as cloing    *
*         : file and re-opening with AppendFile, except works quicker.         *
*         : Return: True = Success, False = Read/write error                   *
********************************************************************************
}   
Public Function SaveFile() As Boolean                                           
Dim TempCurrentSectorPos As Word
Dim TempCurrentSector As LongWord
Dim TempFileSize As LongWord
   If Disk.RWError Then
      SaveFile = False
      Exit
   EndIf
   TempCurrentSectorPos = File(1).CurrentSectorPos   // Store pointers to temp pointers
   TempCurrentSector = File(1).CurrentSector
   TempFileSize = File(1).Size
   CloseFile()                                       // Save file size etc..
   File(1).IsOpen = True
   File(1).CurrentSector = TempCurrentSector         // Recover pointers from temp pointers
   ReadSector(File(1).CurrentSector)                 // Recover buffer contents
   File(1).CurrentSectorPos = TempCurrentSectorPos
   File(1).Size = TempFileSize
   SaveFile = True
End Function
{
********************************************************************************
* Name    : FileSize (PUBLIC)                                                  *
* Purpose : Find size of file (in Bytes)                                       *
********************************************************************************
}   
Public Function FileSize() As LongWord
   FileSize = File(1).Size
End Function
{
********************************************************************************
* Name    : RmDir (OVERLOAD)                                                   *
* Purpose : Remove an existing directory.                                      *
*         : Return: errOK = Success, errNotFound = Directory not found,        *
*         : errDirNotEmpty = Directory not empty, errRWError = Read/write      *
*         : error                                                              *
********************************************************************************
}   
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then                                      
Public Function RmDir(pDirName As TSDName) As Byte
Dim TempName As TSDName
   If Disk.RWError Then
      RmDir = errRWError
      Exit
   EndIf
   If Not(ChDir(pDirName)) Then                // Change into directory
      RmDir = errNotFound                      // Error - directory does not exist
      Exit                                                                    
   EndIf
   Dir(TempName, dirFirst, sdFile)
   If (TempName.Name <> Null) Or (TempName.Extension <> Null) Then
      RmDir = errDirNotEmpty                   // Error - directory is not empty - file(s) exists
      ChDir(cdUp)
      Exit                                                                    
   EndIf
   Dir(TempName, dirFirst, sdDirectory)
   Repeat
      If (TempName.Name <> Null) And (TempName.Name <> ".") And (TempName.Name <> "..") Then
         RmDir = errDirNotEmpty                // Error - directory is not empty - sub-directory(s) exists
         ChDir(cdUp)
         Exit                                                                    
      EndIf
      Dir(TempName, dirNext, sdDirectory)
   Until TempName.Name = Null
   ChDir(cdUp)
   Delete(pDirName, False)
   RmDir = errOK               // No need to check if directory not found on delete, since already checked
End Function
#endif
{
********************************************************************************
* Name    : RmDir (OVERLOAD)                                                   *
* Purpose : Remove an existing directory.                                      *
*         : Return: errOK = Success, errNotFound = Directory Not found,        *
*         : errDirNotEmpty = Directory Not empty, errRWError = Read/Write      *
*         : error                                                              *
********************************************************************************
}   
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then
Public Function RmDir(pDirName As String) As Byte
   Result = RmDir(StrToFileDirName(pDirName))
End Function
#endif
{
********************************************************************************
* Name    : Rename (OVERLOAD)                                                  *
* Purpose : Rename a file from pOldName to pNewName.                           *
*         : pFileDir: sdFile = renaming file, sdDirectory = renaming directory *
*         : Return: errOK = Success, errExists = new file/dir name already     *
*         : exists, errNotFound = old file/dir name not found,                 *
*         : errRWError = Read/write error                                      *
********************************************************************************
}   
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then                                      
Public Function Rename(pOldName As TSDName, pNewName As TSDName, pFileDir As Bit = sdFile) As Byte
#else
Public Function Rename(pOldName As TSDName, pNewName As TSDName) As Byte
#endif
   If Disk.RWError Then
      Rename = errRWError
      Exit
   EndIf
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then
   If pFileDir = sdDirectory Then
      pOldName.Extension = ""                             // Clear extension if directory
      pNewName.Extension = ""
   EndIf
#endif
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then
   If FindRootDirEntry(pNewName, 1, pFileDir) Then        // Check new name does not already exist
#else
   If FindRootDirEntry(pNewName, 1, sdFile) Then
#endif
      Rename = errExists
      Exit
   EndIf
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then
   If Not(FindRootDirEntry(pOldName, 1, pFileDir)) Then   // Check old name does exist;
                                                          // also loads File.FileEntryDirectorySector & Pos
#else
   If Not(FindRootDirEntry(pOldName, 1, sdFile)) Then
#endif
      Rename = errNotFound
      Exit
   EndIf
   ReadSector(File(1).FileEntryDirectorySector)
   InsertFileNameIntoBuffer(pNewName)                     // Overwrite with new name
   WriteSector(File(1).FileEntryDirectorySector)
   Rename = errOK
End Function
{
********************************************************************************
* Name    : Rename (OVERLOAD)                                                  *
* Purpose : Rename a file from pOldName to pNewName.                           *
*         : pFileDir: sdFile = renaming file, sdDirectory = renaming directory *
*         : Return: errOK = Success, errExists = New file/dir name already     *
*         : exists, errNotFound = old File/Dir Name Not found,                 *
*         : errRWError = Read/Write error                                      *
********************************************************************************
}   
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then
Public Function Rename(pOldName As String, pNewName As String, pFileDir As Bit = sdFile) As Byte
#else
Public Function Rename(pOldName As String, pNewName As String) As Byte
#endif
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then
   Result = Rename(StrToFileDirName(pOldName), StrToFileDirName(pNewName), pFileDir)
#else
   Result = Rename(StrToFileDirName(pOldName), StrToFileDirName(pNewName))
#endif
End Function
{
********************************************************************************
* Name    : DiskMounted (PUBLIC)                                               *
* Purpose : Check if disk is mounted - must initialise disk first with Init()  *
*         : before disk will return DiskMounted as True; removing disk after   *
*         : this will return DiskMounted as False.                             *
*         : Return: True = Disk mounted, False = Disk not mounted              *
********************************************************************************
}
Public Function DiskMounted() As Boolean
   ReadSector(0, False)
   DiskMounted = Not(Disk.RWError)
End Function
{
********************************************************************************
* Name    : RWError (PUBLIC)                                                   *
* Purpose : Determine if a disk read/write error has occured.                  *
*         : Return: False = No error occurred, True = Error occurred, only     *
*         : cleared by Init()                                                  *
********************************************************************************
}   
Public Function RWError() As Boolean
   Result = Disk.RWError
End Function   
{
********************************************************************************
* Name    : IsOpen (PUBLIC)                                                    *
* Purpose : Determine if a file is in use.                                     *
*         : Return: True = File in use, False = File not in use                *
********************************************************************************
}   
Public Function IsOpen() As Boolean
   Result = File(1).IsOpen
End Function   
{
********************************************************************************
* Name    : FilePtr (PUBLIC)                                                   *
* Purpose : Return the current file pointer position                           *
********************************************************************************
}   
Public Function FilePtr() As LongWord
   Result = File(1).BytesRead
End Function   
{
********************************************************************************
* Name    : MemoryType (PUBLIC)                                                *
* Purpose : Return the current memory card type.                               *
*         : Return: sdMMC = Multi Media Card, sdSD = Secure Digital Card,      *
*         : sdSDHC = High Capacity Secure Digital Card                         *
********************************************************************************
}   
Public Function MemoryType() As Byte
   Result = Disk.CardType
End Function
{
********************************************************************************
* Name    : SerialNumber (PUBLIC)                                              *
* Purpose : Return the current memory card serial number                       *
********************************************************************************
}   
Public Function SerialNumber() As LongWord
Dim TimeOut, Error As Byte
Dim Index As Word
Dim CID(16) As Byte
#if SD_REINIT_SPI = TRUE
   ReInitSPIPort()
#endif
   SerialNumber = 0
   TimeOut = $00
   Error = 1
   Repeat
      CS = 0                                     
      If SendCmd(CMD10, $00) = $00 Then          // Send Cmd 10
         If SeekResponse($FE) = $FE Then         // Read start token
            Index = 16
            Repeat
               Dec(Index)
               CID(Index) = ReceiveByte()
            Until Index = 0
            ReceiveByte()                        // Read dummy CRC to conclude data block
            ReceiveByte()
            SerialNumber.Byte3 = CID(6)
            SerialNumber.Byte2 = CID(5)
            SerialNumber.Byte1 = CID(4)
            SerialNumber.Byte0 = CID(3)
            Error = 0                            // Read completed successfully
         EndIf   
      EndIf      
      CS = 1
      SendByte($FF)                              // Clock SD/MMC to complete read
      Inc(TimeOut)
   Until Error = 0 Or TimeOut > $01
End Function
{
********************************************************************************
* Name    : FATType (PUBLIC)                                                   *
* Purpose : Return the current memory card FAT type.                           *
*         : Return: sdFAT16 = FAT16, sdFAT32 = FAT32                           *
********************************************************************************
}   
Public Function FATType() As Byte
   Result = Disk.FATType
End Function
{
********************************************************************************
* Name    : WriteBootSectorString (PRIVATE)                                    *
* Purpose : Write section of BootSector                                        *
********************************************************************************
}   
Sub WriteBootSectorString(pVolumeID As LongWord, pOffset As Byte)
   FSR0 = AddressOf(Shared.CurrentSectorBuffer(pOffset))
   POSTINC0 = $29              // Boot Signature
   POSTINC0 = pVolumeID.Byte0
   POSTINC0 = pVolumeID.Byte1
   POSTINC0 = pVolumeID.Byte2
   POSTINC0 = pVolumeID.Byte3
   POSTINC0 = "N"
   POSTINC0 = "0"
   POSTINC0 = " "
   POSTINC0 = "N"
   POSTINC0 = "A"
   POSTINC0 = "M"
   POSTINC0 = "E"
   POSTINC0 = " "
   POSTINC0 = " "
   POSTINC0 = " "
   POSTINC0 = " "
   POSTINC0 = "F"
   POSTINC0 = "A"
   POSTINC0 = "T"
   POSTINC0 = "1"
   POSTINC0 = "6"
   POSTINC0 = " "
   POSTINC0 = " "
   POSTINC0 = " "     
End Sub
{
********************************************************************************
* Name    : Format (PUBLIC)                                                    *
* Purpose : Format SD/MMC card.                                                *
*         : Will delete all files, and re-build the file structure.            *
*         : Return: errOK = Success, errNoResponse = No response from card,    *
*         : errRWError = Read/write error                                      *
********************************************************************************
}   
Public Function Format(pVolumeID As LongWord) As Byte
Dim TimeOut, Error As Byte
Dim Index As Word
Dim CSD(16) As Byte
Dim CSize As LongWord
Dim CSizeMult As Byte
Dim ReadBlockLen As Byte
Dim BlockLen As Byte
Dim DiskSize As LongWord
Dim ReservedSectors As Word
Dim NumberOfRootEntries As Word
Dim Temp1 As LongWord
Dim Temp2 As Word
   Disk.RWError = False                          // Clear read/write error flag to allow a re-format after failure
   TimeOut = $00
   Error = 1
   If Not InitSequence() Then                    // Init SD/MMC in case Format is called before a succesfull Init (e.g. corrupt card format)
      Disk.RWError = True
      Result = errNoResponse
      Exit
   EndIf
   Repeat
      CS = 0                                     
      If SendCmd(CMD9, $00) = $00 Then           // Send Cmd 9 (Send CSD)
         If SeekResponse($FE) = $FE Then         // Read start token
            Index = 16
            Repeat
               Dec(Index)
               CSD(Index) = ReceiveByte()
            Until Index = 0
            ReceiveByte()                        // Read dummy CRC to conclude data block
            ReceiveByte()
            If Disk.CardType = sdSDHC Then       // CSD Ver 2
               CSize = (CSD(8) And %00111111) << 16
               CSize = CSize + (CSD(7) << 8)
               CSize = CSize + CSD(6)
               CSizeMult = 8
               ReadBlockLen = 9           
            Else                                 // CSD Ver 1
               CSize = (CSD(9) And %00000011) << 10
               CSize = CSize + (CSD(8) << 2)
               CSize = CSize + (CSD(7) >> 6)
               CSizeMult = (CSD(6) And %00000011) << 1 
               CSizeMult = CSizeMult + (CSD(5) >> 7)
               ReadBlockLen = CSD(10) And %00001111
            EndIf             
            BlockLen = 1 << (ReadBlockLen - 9)   // Block length in 512 byte sectors
            DiskSize = (CSize + 1) * (1 << (CSizeMult + 2)) * BlockLen   // Memory capacity in 512 byte sectors
            Error = 0                            // Read completed successfully
         EndIf   
      EndIf      
      CS = 1
      SendByte($FF)                              // Clock SD/MMC to complete read
      Inc(TimeOut)
   Until Error = 0 Or TimeOut > $01
   If Error <> 0 Then
      Disk.RWError = True
      Result = errNoResponse
      Exit
   EndIf
   Select DiskSize
      Case < 8401        Disk.SectorsPerCluster = 0     // FAT16 disks up to 4.1 MB, the 0 value should identifies an error
      Case < 32681       Disk.SectorsPerCluster = 2     // FAT16 disks up to 16 MB, 1k cluster
      Case < 262145      Disk.SectorsPerCluster = 4     // FAT16 disks up to 128 MB, 2k cluster
      Case < 524289      Disk.SectorsPerCluster = 8     // FAT16 disks up to 256 MB, 4k cluster
      Case < 1048577     Disk.SectorsPerCluster = 16    // FAT16 disks up to 512 MB, 8k cluster
      Case < 16777217    Disk.SectorsPerCluster = 8     // FAT32 disks up to 8 GB, 4k cluster
      Case < 33554433    Disk.SectorsPerCluster = 16    // FAT32 disks up to 16 GB, 8k cluster
      Case < 67108865    Disk.SectorsPerCluster = 32    // FAT32 disks up to 32 GB, 16k cluster
      Case < $FFFFFFFF   Disk.SectorsPerCluster = 64    // FAT32 disks greater than 32GB, 32k cluster
   EndSelect
   If DiskSize < 1048577 Then                           // FAT16
      ReservedSectors = 1
      NumberOfRootEntries = 512
      Disk.FATType = 1
   Else                                                 // FAT32
      ReservedSectors = 32
      NumberOfRootEntries = 0
      Disk.FATType = 2
   EndIf  
   Temp1 = DiskSize - ReservedSectors - (NumberOfRootEntries / 16)
   Temp2 = (256 * Disk.SectorsPerCluster) + 2
   Temp2 = Temp2 / Disk.FATType
   Disk.SectorsPerFAT = (Temp1 + (Temp2 - 1)) / Temp2     
   Clear(Shared.CurrentSectorBuffer)                    // Clear buffer ready for writing Boot Sector
   Shared.CurrentSectorBuffer(0) = $EB
   Shared.CurrentSectorBuffer(2) = $90
   Shared.CurrentSectorBuffer(3) = "M"                  // OEM Name
   Shared.CurrentSectorBuffer(4) = "S"
   Shared.CurrentSectorBuffer(5) = "W"
   Shared.CurrentSectorBuffer(6) = "I"
   Shared.CurrentSectorBuffer(7) = "N"
   Shared.CurrentSectorBuffer(8) = "4"
   Shared.CurrentSectorBuffer(9) = "."
   Shared.CurrentSectorBuffer(10) = "1"
   Shared.CurrentSectorBuffer(12) = %00000010           // MSB of Bytes Per Sector (512)
   Shared.CurrentSectorBuffer(13) = Disk.SectorsPerCluster
   Shared.CurrentSectorBuffer(14) = ReservedSectors.Byte0
   Shared.CurrentSectorBuffer(15) = ReservedSectors.Byte1
   Shared.CurrentSectorBuffer(16) = 2                   // Number of FATs
   Shared.CurrentSectorBuffer(21) = $F8
   If Disk.FATType = 1 Then                             // FAT16
      Shared.CurrentSectorBuffer(18) = %00000010        // MSB of Number of Root Entries (512)
      Shared.CurrentSectorBuffer(22) = Disk.SectorsPerFAT.Byte0
      Shared.CurrentSectorBuffer(23) = Disk.SectorsPerFAT.Byte1
      WriteBootSectorString(pVolumeID, 38)
   Else                                                 // FAT32
      Shared.CurrentSectorBuffer(36) = Disk.SectorsPerFAT.Byte0
      Shared.CurrentSectorBuffer(37) = Disk.SectorsPerFAT.Byte1
      Shared.CurrentSectorBuffer(38) = Disk.SectorsPerFAT.Byte2
      Shared.CurrentSectorBuffer(39) = Disk.SectorsPerFAT.Byte3
      Shared.CurrentSectorBuffer(44) = 2                // Root Cluster
      Shared.CurrentSectorBuffer(48) = 1                // FSInfo @ Sector 1
      Shared.CurrentSectorBuffer(50) = 6                // BU of Boot Sector @ Sector 6
      WriteBootSectorString(pVolumeID, 66)
      Shared.CurrentSectorBuffer(85) = "3"              // Overwrite FAT16 with FAT32
      Shared.CurrentSectorBuffer(86) = "2"
   EndIf
   If (Disk.FATType = 1) And (DiskSize < $10000) Then   // Total sector count fits in BPB_TotSec16
      Shared.CurrentSectorBuffer(19) = DiskSize.Byte0
      Shared.CurrentSectorBuffer(20) = DiskSize.Byte1     
   Else                                                 // Total sector count must go in BPB_TotSec32
      Shared.CurrentSectorBuffer(32) = DiskSize.Byte0
      Shared.CurrentSectorBuffer(33) = DiskSize.Byte1     
      Shared.CurrentSectorBuffer(34) = DiskSize.Byte2
      Shared.CurrentSectorBuffer(35) = DiskSize.Byte3     
   EndIf  
   Shared.CurrentSectorBuffer(510) = $55
   Shared.CurrentSectorBuffer(511) = $AA
   WriteSector(0)                                       // Write Boot Sector
   If Disk.FATType = 2 Then                             // FAT32
      WriteSector(6)                                    // Write BU of Boot Sector
      Clear(Shared.CurrentSectorBuffer)                 // Clear buffer ready for writing FSInfo
      Shared.CurrentSectorBuffer(0) = $52               // FSInfo lead signature
      Shared.CurrentSectorBuffer(1) = $52
      Shared.CurrentSectorBuffer(2) = $61                  
      Shared.CurrentSectorBuffer(3) = $41
      Shared.CurrentSectorBuffer(484) = $72             // 2nd signature
      Shared.CurrentSectorBuffer(485) = $72
      Shared.CurrentSectorBuffer(486) = $41
      Shared.CurrentSectorBuffer(487) = $61
      Shared.CurrentSectorBuffer(488) = $FF             // Free cluster count (not used)
      Shared.CurrentSectorBuffer(489) = $FF
      Shared.CurrentSectorBuffer(490) = $FF
      Shared.CurrentSectorBuffer(491) = $FF
      Shared.CurrentSectorBuffer(492) = $FF             // Next free cluster (not used)
      Shared.CurrentSectorBuffer(493) = $FF
      Shared.CurrentSectorBuffer(494) = $FF
      Shared.CurrentSectorBuffer(495) = $FF
      Shared.CurrentSectorBuffer(510) = $55             // Trail signature
      Shared.CurrentSectorBuffer(511) = $AA
      WriteSector(1)
      WriteSector(7)
      Clear(Shared.CurrentSectorBuffer)                 // Clear buffer ready for writing empty sector & signature to Sectors 2 & 8
      Shared.CurrentSectorBuffer(510) = $55             // Trail signature
      Shared.CurrentSectorBuffer(511) = $AA
      WriteSector(2)
      WriteSector(8)
   EndIf
   If Disk.RWError = False Then                         // Check no write errors in writing boot sectors
      Result = Init()
      If Result = errOK Then
         If QuickFormat() Then
            Result = errOK
         Else
            Result = errRWError
         EndIf
      EndIf
   Else
      Result = errRWError
   EndIf
End Function
{
********************************************************************************
* Name    : DestroyFormat (PUBLIC)                                             *
* Purpose : Destroy the format of the SD/MMC card by wiping the first          *
*         : pNumberOfSectors sectors.                                          *
*         : WARNING: this routine will delete all files, and destroy the       *
*         : formatting of the card. It will render it unusable without a full  *
*         : format using the Format command. USE WITH CARE! You have been      *
*         : warned! Note that this command will also render the card unusable  *
*         : in a PC until the Format command is run on it to re-instate a file *
*         : system.                                                            *
*         : Return: True = Format successfully destroyed, False = Fail         *
********************************************************************************
}   
Public Function DestroyFormat(pNumberOfSectors As Word = 512) As Boolean
   Result = False
   Disk.RWError = False
   If Not InitSequence() Then   // Init SD/MMC in case DestroyFormat is called before a succesfull Init (e.g. corrupt card format)
      Disk.RWError = True
      Exit
   EndIf
   Clear(Shared.CurrentSectorBuffer)   
   While pNumberOfSectors > 0
      Dec(pNumberOfSectors)
      WriteSector(pNumberOfSectors)
   Wend
   If Disk.RWError = False Then
      Result = True
   EndIf
End Function
{
********************************************************************************
* Name    : FileDirCount (PRIVATE)                                             *
* Purpose : Return the number of files or subdirectories in the current        *
*         : directory                                                          *
********************************************************************************
}   
Function FileDirCount(pFileDir As Bit) As Word
Dim FileName As TSDName
Dim TempFileNumber As Integer
   TempFileNumber = File(1).Number       // Save current file number for Dir
   Result = 0
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then                                      
   SD.Dir(FileName, dirFirst, pFileDir)
#else
   SD.Dir(FileName, dirFirst)
#endif
   While FileName.Name <> Null
      Inc(Result)
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then                                      
      SD.Dir(FileName, dirNext, pFileDir)
#else
      SD.Dir(FileName, dirNext)
#endif
   Wend
   File(1).Number = TempFileNumber       // Restore current file number
End Function
{
********************************************************************************
* Name    : FileCount (PUBLIC)                                                 *
* Purpose : Return the number of files in the current directory                *
********************************************************************************
}   
Public Function FileCount() As Word
   Result = FileDirCount(sdFile)
End Function
{
********************************************************************************
* Name    : DirCount (PUBLIC)                                                  *
* Purpose : Return the number of subdirectories in the current directory       *
********************************************************************************
}   
#if SD_SUPPORT_SUB_DIRECTORIES = TRUE Then                                      
Public Function DirCount() As Word
   Result = FileDirCount(sdDirectory)
End Function
#endif


// module initialisation...
#if SD_SPI = MSSP
   SSPControl1 = %00100010  // Setup MSSP module at spiOscDiv64 initially
   SSPStatus = %11000000                         
#endif
Disk.RWError = True         // Start module with read/write error to prevent disk access until initialised
